"""
stupeflixbase.py

Base code for the Stupeflix video generation web services.
See stupeflix.py and LICENSE.txt for more information.

http://www.stupeflix.com
http://code.google.com/p/stupeflix/

"""

import os
import hashlib
import base64 
import connection
import time
import hmac
import logging


logger = logging.getLogger(__name__)


class StupeflixBase(object): 
    def __init__(self, accessKey, privateKey, host = "http://services.stupeflix.com", service = 'stupeflix-1.0', debug = False):
        self.accessKey = str(accessKey)
        self.privateKey = str(privateKey)
        self.base_url = host + '/' + service
        self.debug = debug
        self.service = service
        self.TEXT_XML_CONTENT_TYPE = "text/xml"
        self.APPLICATION_ZIP_CONTENT_TYPE = "application/zip"
        self.APPLICATION_JSON_CONTENT_TYPE = "application/json"
        self.APPLICATION_URLENCODED_CONTENT_TYPE = "application/x-www-form-urlencoded"
        self.PROFILES_PARAMETER = "Profiles"
        self.XML_PARAMETER = "ProfilesXML"
        self.MARKER_PARAMETER = "Marker"
        self.MAXKEYS_PARAMETER = "MaxKeys"
        # Currently there is only the Marker parameter (used for partial enumeration)
        self.parametersToAdd = [self.MARKER_PARAMETER, self.MAXKEYS_PARAMETER]
        # Time between two retry (to wait for Amazon S3)
        self.sleepTime = 5.0
        self.maxRetry = 5
        self.base = True

    def connectionGet(self, base_url = None):
        if base_url == None:
            base_url = self.base_url
        return connection.Connection(base_url)

    def paramString(self, parameters):
        paramStr = ""
        if parameters:
            for p in self.parametersToAdd:
                if p in parameters:
                    paramStr += "%s\n%s\n" % (p, parameters[p])


        if isinstance(paramStr, unicode):
            paramStr = paramStr.encode("utf8")

        return paramStr

    def strToSign(self, method, resource, md5, mime, datestr, parameters):
        paramStr = self.paramString(parameters)
        stringToSign  = "%s\n%s\n%s\n%s\n%s\n%s" % (method, md5, mime, datestr, '/' + self.service + resource, paramStr)
        return stringToSign
        
    def sign(self, strToSign, secretKey):               
        h = hmac.new(secretKey, strToSign, hashlib.sha1)
        sig = h.hexdigest()
        return sig

    def signUrl(self, url, method, md5, mime, parameters = {}, inlineAuth = False):
        now = int(time.time())
        strToSign = self.strToSign(method, url, md5, mime, now, parameters)            
        signature = self.sign(strToSign, self.privateKey)
        if inlineAuth:
            url += "%s/%s/%s/" % (self.accessKey, signature, now)
        else:
            url += "?Date=%s&AccessKey=%s&Signature=%s" % (now, self.accessKey,signature)
        if parameters:
            for k,v in parameters.items():
                url += "&%s=%s" % (k,v)
        return url

    def md5FileOrBody(self, filename, body = None):
        md5 = hashlib.md5()
        if body != None:
            md5.update(body)
        else:
            chunksize=1024
            f = open(filename, 'rb')
        
            while True:
                chunk = f.read(chunksize)
                if not chunk: break
                md5.update(chunk)
            f.close()     

        digest = md5.digest()
        return (digest, md5.hexdigest(), base64.b64encode(digest)) 

    def isZip(self, filename):
        f = open(filename, 'rb')
        header = f.read(4)
        return header == r'PK'+chr(3)+chr(4)

    def logdebug(self, s):
        if self.debug:
            print str(s)

    def error(self, message):
        self.logdebug(message)
        raise Exception, message

    def answer_error(self, answer, message):
        raise Exception, "%s\n%s" % (message, answer['body'])

    # sendcallback is an object with 
    #  - a 'sendCallBack' member function that accept a unique int argument (=number of bytes written so far)
    #  - a 'sendBlockSize' member function with no argument which return the size of block to be sent
    def sendContent(self, method, url, contentType, filename = None, body = None,  parameters = None, sendcallback = None, streamBody = None, base_url = None):
        if streamBody == None:
            # fix default ascii encoding
            if body:
                body = body.encode('utf-8')

            md5, md5hex, md5base64 = self.md5FileOrBody(filename, body)

            if filename:
                size = os.stat(filename).st_size 
            else:
                if isinstance(body, unicode):
                    body = body.encode("utf8")
                size = len(body)
        else:
            md5, md5hex, md5base64 = streamBody.md5Triplet()
            body = streamBody
            size = None

        # SEND DATA
        conn = self.connectionGet(base_url)

        headers = {'Content-MD5':str(md5base64), 
                   'Content-Type':contentType}

        if size != None:
            headers['Content-Length'] = str(size)
        else:
            headers['Transfer-Encoding'] = 'chunked'

        url = self.signUrl(url, method, md5base64, contentType, parameters)

        # LAUNCH THE REQUEST : TODO : pass filename instead of body 
        if method == "PUT":
            answer = conn.request_put(url, filename = filename, body = body, headers = headers, sendcallback = sendcallback)
        elif method == "POST":
            answer = conn.request_post(url, filename = filename, body = body, headers = headers)
        elif method == "DELETE":
            answer = conn.request_delete(url, headers = headers)

        headers = answer['headers']

        self.logdebug(headers)
        self.logdebug(answer['body'])

        # NOW CHECK THAT EVERYTHING IS OK
        status = headers['status']
        if status != '200':
            msg = "sendContent : bad STATUS %s" % status
            self.answer_error(answer, msg)
        
        if 'etag' not in headers:
            msg = "corrupted answer: no etag in headers. Response body is " + answer['body']
            self.error(msg)
            
        obtainedMD5 = headers['etag'].strip('"')         

        if obtainedMD5 != md5hex:
            msg = "sendContent : bad returned etags %s =! %s (ref)" % (obtainedMD5, md5hex)
            self.error(msg)

        return answer

    def getContentUrl(self, url, method, parameters, inlineAuth = False):
        return self.signUrl(url, method, "", "", parameters, inlineAuth  = inlineAuth)

    def getContent(self, url, filename = None, parameters = None):
        sleepTime = self.sleepTime
        for i in range(self.maxRetry):
            raiseExceptionOn404 = (i + 1) == self.maxRetry
            ret = self.getContent_(url, filename, parameters, raiseExceptionOn404)
            if ret["status"] != 404:
                return ret
            elif ret["status"] >= 300 and ret["status"] < 400:
                url = ret["url"]            

            logger.info("RETRYING: %s", i)
            time.sleep(sleepTime)

    def getContent_(self, url, filename = None, parameters = None, raiseExceptionOn404 = True):

        method = "GET"
        url = self.getContentUrl(url, method, parameters)

        # GET DATA
        conn = self.connectionGet()
        answer = conn.request_get(url)
        body = answer['body']

        headers = answer['headers']
        status = int(headers['status'])

        ret = {"status":status}
        if 'content-location' in headers:
            ret['url'] = headers['content-location']

        if 'content-length' in headers:
            obtainedSize = int(headers['content-length'])

        if status == 204:
            # there was no content
            obtainedSize = 0
            if len(body) != 0:
                self.error("204 status with non empty body.")
        elif status == 200:
            pass
        elif status == 404 and not raiseExceptionOn404:            
            return ret
        elif status >= 300 and status < 400:
            return ret
        else:
            msg = "getContent, url=%s%s : bad STATUS %s" % (self.base_url, url, status)
            self.answer_error(answer, msg)

        if len(body) != obtainedSize:
            self.error("Non matching body length and content-length")

        if filename != None:
            f = open(filename, 'wb')
            f.write(body)            
            f.close()

            if status == 204:
                os.unlink(filename)
            else:
                filesize = os.stat(filename).st_size             
                if obtainedSize != filesize:
                    os.unlink(filename)
                    self.error("file size is incorrect : file size = %d, body size = %d" % (filesize, obtainedSize))            
                
        # NOW CHECK EVERYTHING IS OK
        md5, md5hex, md5base64 = self.md5FileOrBody(filename, body)

        if status != 204:
            obtainedMD5 = headers['etag'].strip('"') 
            if obtainedMD5 != md5hex:
                if filename:
                    os.unlink(filename)
                    pass
                self.error("getContent : bad returned etag %s =! %s (ref)" % (md5hex, obtainedMD5))

        if status == 200:
            #self.logdebug("headers = %s" % headers)
            ret = {'size':obtainedSize, 'url':url, 'headers': headers}
            if 'content-location' in headers:
                ret['url'] = headers['content-location']
        else:
            ret = {'size':obtainedSize, 'url':url}
        
        if not filename:
            ret['body'] = body

        ret["status"] = status

        return ret




try:
    import simplejson as json
except Exception, e:
    import json
import urllib


class HTTPError(Exception):
    def __init__(self, code, message):
        super(HTTPError, self).__init__("HTTPError %s: %s" % (code, message))


class LizardClient(object):
    def __init__(self,  baseurl, accesskey = None, secret = None, auth_model = "secret"):
        self.auth_model = auth_model
        self.accesskey = accesskey 
        self.secret = secret
        self.token = None
        self.baseurl = baseurl
        self.session = None

        if auth_model == "secret":
            if self.secret == None:
                raise Exception, "Secret key was not specified."
            if not self.baseurl.startswith("https://"):
                raise Exception, "Secret based auth can only be used with https"
        
    def streamingGet(self, url, body):
        line = ""
        opened = urllib.urlopen(url, body)
        while True:           
            c = opened.read(1)
            if c != "\n" and len(c) != 0:
                line += c
                continue                                       
            yield line
            line = ""
            if len(c) == 0:
                break

    def streamingJSONGet(self, url, body = None):
        for line in self.streamingGet(url, body):
            try:
                payload = json.loads(line)
            except ValueError, e:
                raise ValueError, "Could not decode json from line %s" % line
            yield payload    

    def taskPrepare_(self, response = "status", **kwargs):
        parameters = {}
        
        for k,v in kwargs.items():
            if v == None:
                continue
            if isinstance(v, str) or isinstance(v, unicode):
                parameters[k] = v.encode("utf8")
            elif isinstance(v, float) or isinstance(v, int):
                parameters[k] = str(v)
            else:
                parameters[k] = json.dumps(v)

        parameters["response"] = response
        return parameters
            
    def authAdd(self, parameters, auth_model = None):
        if auth_model == None:
            auth_model = self.auth_model
        if auth_model == "secret":
            parameters["secret"] = self.secret
        elif auth_model == "public":
            parameters["accesskey"] = self.accesskey

    def taskDoURL(self, taskType = None, taskkey = None, filekey = None, response = "result", method = "GET", auth_model = None, https = True, **kwargs):
        assert(taskType != None or taskkey != None)
        key0 = taskType if taskType != None else taskkey
        if filekey == None:            
            url = "%s/%s/" % (self.baseurl, key0)
        else:
            url = "%s/%s/%s" % (self.baseurl, key0, filekey)
        parameters = self.taskPrepare_(response = response, **kwargs)

        self.authAdd(parameters, auth_model = auth_model)

        if not https:
            if url.startswith("https://"):
                url = "http" + url[5:]
        else:
            if url.startswith("http://"):
                url = "https" + url[4:]                                

        parameters = urllib.urlencode(parameters)
        if method == "GET":
            return url + "?" + parameters, None
        elif method == "POST":
            return  url, parameters
        else:
            raise Exception, "Invalid method"

    # Return meta, content
    def taskDo(self, taskType = None, taskkey = None, filekey = None,  response = "result", method = "GET", **kwargs):
        url, body = self.taskDoURL(taskType = taskType, taskkey = taskkey, filekey = filekey, response = response, method = method, **kwargs)
#        print url, body
        if response in ["result"]:
            return urllib.urlopen(url, body)
        elif response in ["status", "status_final"]:
            # Return single object
            u = urllib.urlopen(url, body)            
            body = u.read()
            code = 200
            try:
                code = u.getcode()
            except Exception, e:
                pass
            if code >= 400 and code < 600:
                raise HTTPError(code, body)
            try:
                ret = json.loads(body)
                return ret
            except Exception, e:
                raise Exception, "Could not decode json from %s" % body
        elif response in ["status_stream"]:
            # Return generator suitable for "for x in " loop
            return self.streamingJSONGet(url, body)
        else:
            raise Exception, "Invalid response type %s, should be in result, status, status_final, status_streaming" % response



"""
Stupeflix web services python client : stupeflix.py

This is the main entry point for using stupeflix video generation web service.

See README for more information

http://www.stupeflix.com
http://code.google.com/p/stupeflix/

Copyright (c) 2008 Francois Lagunas, http://www.stupeflix.com

"""

import urllib
from stupeflixbase import StupeflixBase
import demjson
import xml.sax.saxutils as saxutils
import connection
import types

# TODO : file streaming

class StupeflixClient(StupeflixBase): 
    def __init__(self, accessKey, privateKey, host = "http://services.stupeflix.com", service = 'stupeflix-1.0', debug = False):
        if host.endswith("/"):
            host = host[:-1]
        super(StupeflixClient, self).__init__(accessKey, privateKey, host, service, debug)
        self.batch = False
        self.batchData = ""

    # Start a batch, used for speeduping video definition upload
    # Operation that can be batched : sendDefinition and createProfiles
    # Operation 
    # Only works for xml definition, not zip, and xml must be in UTF8
    def batchStart(self, maxSize = 1000000):
        self.batch = True
        self.batchData = "<batch>"
        self.batchMaxSize = maxSize

    # End a batch: actually send data
    def batchEnd(self):
        self.batchData += "</batch>"
        self.sendDefinitionBatch(body = self.batchData)
        self.batchData = ""
        self.batch = False

    # Send a definition file to the API
    def sendDefinition(self, user, resource, filename = None, body = None):
        url = self.definitionUrl(user, resource)
        if body:
            contentType = self.TEXT_XML_CONTENT_TYPE;
        elif self.isZip(filename):
            contentType = self.APPLICATION_ZIP_CONTENT_TYPE
        else:
            contentType = self.TEXT_XML_CONTENT_TYPE
        if self.batch and contentType == self.TEXT_XML_CONTENT_TYPE:
            self.batchData += "<task user=\"%s\" resource=\"%s\">" % (user, resource)
            if body:                
                self.batchData += body
            else:
                self.batchData += open(filename, 'rb').read()
        else:
            return self.sendContent("PUT", url, contentType, filename, body)

    # Send a definition file to the API
    def sendDefinitionBatch(self, filename = None, body = None):
        url = self.definitionBatchUrl()
        contentType = self.TEXT_XML_CONTENT_TYPE;
        return self.sendContent("PUT", url, contentType, filename, body)

    def getDefinition(self, user, resource, filename):
        url = self.definitionUrl(user, resource)
        return self.getContent(url, filename)['size']
    
    def _getAbsoluteUrl(self, url, followRedirect = False):
        urlPart = self.getContentUrl(url, 'GET', None)
        if followRedirect:
            conn = connection.Connection(self.base_url, followRedirect = False)
            response = conn.request_get(urlPart)
            if "location" in response["headers"]:
                return response["headers"]["location"]
            else:
                return self.base_url + urlPart
        else:
            return self.base_url + urlPart
      
    def getDefinitionUrl(self, user, resource):
        url = self.definitionUrl(user, resource)
        return self._getAbsoluteUrl(url)
      
    def getProfileUrl(self, user, resource, profile, followRedirect = False):
        url = self.profileUrl(user, resource, profile)
        return self._getAbsoluteUrl(url, followRedirect)

    def getProfile(self, user, resource, profile, filename):
        url = self.profileUrl(user, resource, profile)
        return self.getContent(url, filename)['size']

    def getProfileThumbUrl(self, user, resource, profile, followRedirect = False):
        url = self.profileThumbUrl(user, resource, profile, "thumb.jpg")
        return self._getAbsoluteUrl(url, followRedirect)

    def getProfileThumb(self, user, resource, profile, filename):
        url = self.profileThumbUrl(user, resource, profile, "thumb.jpg")
        return self.getContent(url, filename)['size']

    def getProfileReportUrl(self, user, resource, profile, followRedirect = False):
        url = self.profileReportUrl(user, resource, profile)
        return self._getAbsoluteUrl(url, followRedirect)

    def getProfileReport(self, user, resource, profile, filename):
        url = self.profileReportUrl(user, resource, profile)
        return self.getContent(url, filename)['size']

    def getProfilePreviewUrl(self, user, resource, profile):
        url = self.profilePreviewUrl(user, resource, profile)
        return self._getAbsoluteUrl(url, False)
    
    def createProfiles(self, user, resource, profiles):
        profileData = ""
        if isinstance(profiles, list):
            profileData = "<profiles>"
            for p in profiles:
                profileData += "<profile name=\"%s\"><stupeflixStore/></profile>" % p
            profileData += "</profiles>"
        elif isinstance(profiles, StupeflixProfileSet):
            profileData = profiles.xmlGet()
        else:
            profileData = profiles
        if self.batch:
            self.batchData += profileData
            self.batchData += "</task>"
            if len(self.batchData) >= self.batchMaxSize:        
                try:
                    self.batchEnd()
                finally:                    
                    self.batchStart(self.batchMaxSize)
        else:
            url, parameters = self.profileCreateUrl(user, resource, profileData)
            contentType = self.APPLICATION_URLENCODED_CONTENT_TYPE
            body = urllib.urlencode(parameters)

            return self.sendContent("POST", url, contentType, body = body)

    def cancelProfiles(self, user, resource, profiles):
        url, parameters = self.profileCancelUrl(user, resource, profileData)
        contentType = self.APPLICATION_URLENCODED_CONTENT_TYPE
        body = urllib.urlencode(parameters)

        return self.sendContent("DELETE", url, contentType, body = body)

    def getStatus(self, user = None, resource = None, profile = None, marker = None, maxKeys = None):
        url, parameters = self.statusUrl(user, resource, profile, marker, maxKeys)
        ret = self.getContent(url, parameters = parameters)
        status = demjson.decode(ret['body'], encoding='utf-8' )    
        return status

    def getMarker(self, status):
        if len(status) == 0:
            return None
        lastStatus = status[-1]
        return map(lambda x: lastStatus[x], ["user", "resource", "profile"])

    # helper functions : build non signed urls for each kind of action
    def definitionUrl(self, user, resource):
        return "/%s/%s/definition/" % (user, resource)

    # helper functions : build non signed urls for each kind of action
    def definitionBatchUrl(self):
        return "/batch/"

    def profileUrl(self, user, resource, profile):
        return "/%s/%s/%s/" % (user, resource, profile)

    def profileThumbUrl(self, user, resource, profile, thumbname):
        return "/%s/%s/%s/%s/" % (user, resource, profile, thumbname)

    def profileReportUrl(self, user, resource, profile):
        return "/%s/%s/%s/%s/" % (user, resource, profile, "report.xml")

    def profilePreviewUrl(self, user, resource, profile):
        return "/%s/%s/%s/%s/" % (user, resource, profile, "preview.flv")

    def profileCreateUrl(self, user, resource, profiles):
        s = "/%s/%s/" % (user, resource)
        parameters = {self.XML_PARAMETER:profiles.encode("utf8")}
        return s, parameters
    
    def actionUrl(self, user, resource, profile, action):
        path = [user, resource, profile]
        s = ""
        for p in path:
            if p == None:
                break
            s += "/%s" % p
        s += "/%s/" % action            
        return s

    def statusUrl(self, user, resource, profile, marker = None, maxKeys = None):
        params = {}
        if marker != None:
            params[self.MARKER_PARAMETER] = "/".join(marker)
        if maxKeys != None:
            params[self.MAXKEYS_PARAMETER] = maxKeys

        return self.actionUrl(user, resource, profile, "status"), params

class StupeflixXMLNode(object):
    def __init__(self, nodeName, attributes = None, children = None, text = None):
        self.children = children
        self.attributes = attributes
        self.nodeName = nodeName
        self.text = text

    def xmlGet(self):
        docXML = '<' + self.nodeName 
        if self.attributes and len(self.attributes) != 0:
            for k,v in self.attributes.iteritems():
                docXML += " "
                if v == None:
                    v = ""
                k = str(k)
                v = str(v)
                docXML += k + '=' + saxutils.quoteattr(v)
        docXML += '>'
        if self.children:
            for c in self.children:
                docXML += c.xmlGet()
        if self.text:
            docXML += saxutils.escape(self.text)
        docXML += '</' + self.nodeName + '>'
        
        return docXML

    def metaChildrenAppend(self, meta = None, notify = None, children = None):
        childrenArray = []
        if meta:
            childrenArray += [meta]
        if notify:
            childrenArray += [notify]
        if children:
            childrenArray += children
        return childrenArray

class StupeflixMeta(StupeflixXMLNode):
    def __init__(self, dict):
        children = []
        for k,v in dict.iteritems():
            children += [StupeflixXMLNode(k, None, None, v)]
        super(StupeflixMeta, self).__init__("meta", {}, children)

class StupeflixProfileSet(StupeflixXMLNode):
    def __init__(self, profiles, meta = None, notify = None):
        children = self.metaChildrenAppend(meta, notify, profiles)
        super(StupeflixProfileSet, self).__init__("profiles", {}, children)

    @staticmethod
    def deflt(profiles):
        profSet = []
        for p in profiles:
            upload = StupeflixDefaultUpload()
            profSet += [StupeflixProfile(p, [upload])]
            
        return StupeflixProfileSet(profSet)

class StupeflixProfile(StupeflixXMLNode):
    def __init__(self, profileName, uploads = None, meta = None, notify = None):
        children = self.metaChildrenAppend(meta, notify, uploads)
        super(StupeflixProfile, self).__init__("profile", {"name":profileName}, children)
        
class StupeflixNotify(StupeflixXMLNode):
    def __init__(self, url, statusRegexp):
        super(StupeflixNotify, self).__init__("notify", {"url":url, "statusRegexp":statusRegexp})

class StupeflixHttpHeader(StupeflixXMLNode):
    def __init__(self, key, value):
        super(StupeflixHttpHeader, self).__init__("header", {"key":key, "value":value})
            
class StupeflixUpload(StupeflixXMLNode):
    def __init__(self, name, parameters, meta = None, notify = None, children = None):
        children = self.metaChildrenAppend(meta, notify, children)
        super(StupeflixUpload, self).__init__(name, parameters, children)           
        
class StupeflixHttpPOSTUpload(StupeflixUpload):
    def __init__(self, url, meta = None, notify = None):
        super(StupeflixHttpPOSTUpload, self).__init__("httpPOST", {"url":url}, meta, notify)

class StupeflixHttpPUTUpload(StupeflixUpload):
    def __init__(self, url, meta = None, notify = None, headers = None):
        super(StupeflixHttpPUTUpload, self).__init__("httpPUT", {"url":url}, meta, notify, headers)
        
class StupeflixYoutubeUpload(StupeflixUpload):
    def __init__(self, login, password, meta = None, notify = None):
        super(StupeflixYoutubeUpload, self).__init__("youtube", {"login":login, "password":password}, meta, notify)

class StupeflixYoutubeTokenUpload(StupeflixUpload):
    def __init__(self, developerKey, token, meta = None, notify = None):
        super(StupeflixYoutubeTokenUpload, self).__init__("youtube", {"developerkey":developerKey, "sid":token}, meta, notify)

class StupeflixYoutubeOAuthUpload(StupeflixUpload):
    def __init__(self, developerKey, oauthconsumerkey, oauthconsumersecret, oauthtoken, oauthtokensecret, meta = None, notify = None):
        super(StupeflixYoutubeOAuthUpload, self).__init__("youtube", 
                                                          {"developerkey":developerKey,
                                                           "oauthconsumerkey" : oauthconsumerkey, 
                                                           "oauthconsumersecret": oauthconsumersecret, 
                                                           "oauthtoken" : oauthtoken, 
                                                           "oauthtokensecret" : oauthtokensecret},
                                                          meta, notify)

class StupeflixBrightcoveUpload(StupeflixUpload):
    def __init__(self, token, reference_id = None, meta = None, notify = None):
        parameters = {"sid":token}
        if reference_id != None:
            parameters["reference_id"] = reference_id
        super(StupeflixBrightcoveUpload, self).__init__("brightcove", parameters, meta, notify)

class StupeflixFacebookTokenUpload(StupeflixUpload):
    def __init__(self, apiKey, secret, token, meta = None, notify = None):
        super(StupeflixFacebookTokenUpload, self).__init__("facebook", {"apikey":apiKey, "secret":secret, "sid":token}, meta, notify)

class StupeflixDailymotionUpload(StupeflixUpload):
    def __init__(self, login, password, sid = None, meta = None, notify = None):
        if not isinstance(sid, (str, unicode, types.NoneType)) or not isinstance(login, (str, unicode, types.NoneType)) or not isinstance(password, (str, unicode, types.NoneType)):
            raise Exception("Bad type for sid = '%s'" % sid)
        if sid != None:
            parameters = {"sid":sid}
        else:
            parameters = {"login":login, "password":password}
        super(StupeflixDailymotionUpload, self).__init__("dailymotion", parameters, meta, notify)

class StupeflixDefaultUpload(StupeflixUpload):
    def __init__(self, meta = None, notify = None):
        children = self.metaChildrenAppend(meta)
        super(StupeflixDefaultUpload, self).__init__("stupeflixStore", {}, meta, notify)

class StupeflixFTPUpload(StupeflixUpload):
    def __init__(self, server, login, password, directory, meta = None, notify = None):
        if not isinstance(server, (str, unicode, types.NoneType)):
            raise Exception("Bad type for server")

        if not isinstance(login, (str, unicode, types.NoneType)):
            raise Exception("Bad type for login")

        if not isinstance(password, (str, unicode, types.NoneType)):
            raise Exception("Bad type for password")

        if not isinstance(directory, (str, unicode, types.NoneType)):
            raise Exception("Bad type for directory")

        parameters = {"login":login, "password":password, "directory":directory, "server":server}
        super(StupeflixFTPUpload, self).__init__("ftp", parameters, meta, notify)
        
class StupeflixS3Upload(StupeflixUpload):
    def __init__(self, bucket, resourcePrefix, accesskey = None, secretkey = None, meta = None, notify = None):
        parameters = {"bucket":bucket, "resourcePrefix":resourcePrefix}
        if accesskey != None: 
            parameters["accesskey"] = accesskey
        if secretkey != None:
            parameters["secretkey"] = secretkey
        super(StupeflixS3Upload, self).__init__("s3", parameters, meta, notify) 



import key
import conf
from stupeflix import *
import time
import traceback

class StupeflixTest:
    def __init__(self):
        # Create the client to access the API
        # You can safely assume that key.stupeflixHost is None
        self.client = StupeflixClient(key.stupeflixAccessKey, key.stupeflixSecretKey, host = key.stupeflixHost)
        
        # Set the name for the resource to be created
        # These names are alphanumerical, and can be set to whatever you want
        self.user = "test"
        self.resource = "resource" + conf.dateString
        # Configuration for s3 retries
        self.s3Retries = 5
        self.s3Wait = 1
        # Set this to true if you want to see the status of videos being generated
        self.debug = True

    def uploadsCreate(self, profileName):
        """Create several upload targets. Choose the ones you want by setting the corresponding variables in conf.py, or modify directly the code."""
        user = self.user
        resource = self.resource
        
        # Array of uploads to be filled in
        uploads = []

        # Default upload creation : will store to the stupeflix s3 bucket
        # This is not mandatory, just a easy way to store temporarily the result of the video generation.
        uploads += [StupeflixDefaultUpload()]

         # YouTube upload creation, if correct information was entered in conf.py
        if conf.youtubeLogin != None:
            # Create sample youtube information
            tags = ",".join(["these","are","my","tags"])
            youtubeInfo = {"title" : "Upload test "  + conf.dateString,
                           "description": "Upload test description" + conf.dateString,
                           "tags":tags,
                           "channels":"Tech",
                           "acl":"public",
                           "location":"49,-3"}

            youtubeMeta = StupeflixMeta(youtubeInfo)
            # There is no currently notification 
            youtubeNotify = None
            uploads += [StupeflixYoutubeUpload(conf.youtubeLogin, conf.youtubePassword, meta = youtubeMeta, notify = youtubeNotify)]

        # YouTube upload creation, if correct information was entered in conf.py
        if conf.youtubeAuthSubToken != None:
            # Create sample youtube information
            tags = ",".join(["these","are","my","tags"])
            youtubeInfo = {"title" : "Upload test "  + conf.dateString,
                           "description": "Upload test description" + conf.dateString,
                           "tags":tags,
                           "channels":"Tech",
                           "acl":"public",
                           "location":"49,-3"}

            youtubeMeta = StupeflixMeta(youtubeInfo)
            # There is no currently notification 
            youtubeNotify = None
            uploads += [StupeflixYoutubeTokenUpload(conf.youtubeDeveloperKey, conf.youtubeAuthSubToken, meta = youtubeMeta, notify = youtubeNotify)]

        # Dailymotion upload creation, if correct information was entered in conf.py
        if conf.dailymotionLogin != None:
            # Create sample dailymotion information
            tags = ",".join(["these","are","my","tags"])
            dailymotionInfo = {"title" : "Upload test "  + conf.dateString,
                               "description": "Upload test description" + conf.dateString,
                               "tags":tags,
                               "channels":"Tech",
                               "acl":"public",
                               "location":"49,-3"}

            dailymotionMeta = StupeflixMeta(dailymotionInfo)
            # There is no currently notification 
            dailymotionNotify = None
            uploads += [StupeflixDailymotionUpload(conf.dailymotionLogin, conf.dailymotionPassword, meta = dailymotionMeta, notify = dailymotionNotify)]

        # Dailymotion upload creation, if correct information was entered in conf.py
        if conf.facebookApiKey != None:
            # Create sample facebook information
            facebookInfo = {"title" : "Upload test "  + conf.dateString,
                            "description": "Upload test description" + conf.dateString}

            facebookMeta = StupeflixMeta(facebookInfo)
            # There is no currently notification 
            facebookNotify = None
            uploads += [StupeflixFacebookTokenUpload(conf.facebookApiKey, conf.facebookApiSecret, conf.facebookToken, meta = facebookMeta, notify = facebookNotify)]
            
        # S3 Upload creation : upload to your own S3 bucket
        if conf.s3AccessKey != None:
            # Create an arbitrary location for the resource
            s3resource = "%s/%s/%s" % (user, resource, profileName)
            # Create s3 upload settings
            uploads += [StupeflixS3Upload(bucket=conf.s3Bucket, resourcePrefix = s3resource, accesskey = conf.s3AccessKey, secretkey = conf.s3SecretKey)]

        #  HTTP Uploads creation : POST and PUT
        if conf.httpUploadPrefix != None:
            # Create http POST upload settings
            postURL = conf.httpUploadPrefix + "post/%s/%s/%s" % (user, resource, profileName)
            uploads += [StupeflixHttpPOSTUpload(postURL)]

            # Create http PUT upload settings
            putURL = conf.httpUploadPrefix + "put/%s/%s/%s" % (user, resource, profileName)
            uploads += [StupeflixHttpPUTUpload(putURL)]

        return uploads
    
    def availableKey(self, rank, suffix = "id"):
        return  "available-" + str(rank) + "-" + suffix

    # Check that all went fine, until every uploads is finished (or went on error)
    def waitForCompletion(self, uploadCount):
        error = False
        # Then wait for the generation to complete
        available = False
        status = None
        error = False
        while not available and not error:
            # Retrieve an array of status for every profiles for user and resource
            status = self.client.getStatus(self.user, self.resource, None)            
            print status
            # Variable to test if every profile is available
            for s in status:
                if self.debug:
                    print s
                available = True
                for id in range(uploadCount):
                    availableKey = self.availableKey(id)
                    availableType = self.availableKey(id, "type")
                    if not availableKey in s["status"]:
                        if self.debug:
                            print  "upload #" + str(id)  + " '" + self.uploads[id].__class__.__name__ + "' not yet ready for profile " + s["profile"]
                        available = False
                        break
                    else:                 
                        if self.debug:
                            print  "upload #" + str(id)  + " '" + s["status"][availableType] + "' ready for profile " + s["profile"]

                if s["status"]["status"] == "error":
                    error = True
                    break
            time.sleep(5)
        # if available if False, that means that an error occurred
        return available, status

    def s3WaitLoop(self):
        """Sometimes we have to wait for s3 to make the content available (this is in the Amazon S3 spec). This function is built to do just that."""
        s3Wait = self.s3Wait
        for i in range(self.s3Retries):
            try:
                yield None
            except Exception, e:
                print traceback.format_exc(e)
                if (i + 1) == self.s3Retries:
                    raise
                else:
                    time.sleep(s3Wait)
                    s3Wait *= 2

    def run(self):
        """This is the main function for creating videos, with main calls to the Stupeflix API."""
        # The set of profiles to be created
        profileNames = ["iphone"] # You can add profiles there :  "flash-small, quicktime, dvd ..."
        profileArray = []
        for profileName in profileNames:
            # Create an array of uploads to be executed after the current profile is generated.
            # All upload types will be the same here, but of course you can specify different targets for different profiles.
            uploads = self.uploadsCreate(profileName)
            self.uploads = uploads
            # Create a new profile.
            profileArray += [StupeflixProfile(profileName, uploads = uploads)]

        # Notification is not configured there : this consists in a series of HTTP POST ping request to your own server.
        # This is much more powerful than polling the API as demonstrated in function waitForCompletion.
        notify = None
        # Uncomment this line if you want to test notification. StatusRegexp is used to filter notification sent to your server.
        # Here, only final "available message" would be sent
        # notify =  StupeflixNotify(url = "http://myserver.com/mypath", statusRegexp = "available")

        # Create the set of profiles to be created
        profiles = StupeflixProfileSet(profileArray, meta = None, notify = notify)

        # This is only used to give proper names to output files (file names are appended with a proper extension) 
        profileExtensions = ["mp4"] # flv ...

        # Calls to the API Start there

        # First send the movie definition file to the service. (see sample movie.xml in this directory)
        self.client.sendDefinition(self.user, self.resource, conf.filename)

        # Then launch the generation, using the configuration we have built earlier
        self.client.createProfiles(self.user, self.resource, profiles)
        
        # Poll the API, waiting for completion
        available, status = self.waitForCompletion(uploadCount = len(uploads))

        # Check if everything went fine
        if not available :
            # Something went bad: at least some part of the task was not complete, but some may still have or even was uploaded.
            # This may happen for example if upload to youtube failed but upload to your own server succeeded.
            # an error occured, the status will give more information
            raise Exception, str(status)

        # Download all profiles : this is only valid if a StupeflixDefaultUpload was used in the uploads array
        for i,p in enumerate(profileNames):
            # Print the profile url were the video can be found
            print self.client.getProfileUrl(self.user, self.resource, p)
            movieName = "movie%s.%s" % (p, profileExtensions[i])
            for n in self.s3WaitLoop():
                self.client.getProfile(self.user, self.resource, p, movieName)
                
            # Download the profile thumb url
            thumbName = "thumb_%s.jpg" % p
            for n in self.s3WaitLoop():
                self.client.getProfileThumb(self.user, self.resource, p, thumbName)                

        print "Test succeeded."


def main():
    # Test if the keys were set
    if key.stupeflixAccessKey == None:
        print "ERROR : Please fill in key information in key.py"
        exit(1)
    test = StupeflixTest()
    test.run()



import os

# These are required variables, by default read from the environment variables, you can put your own directly here. Register at http://www.stupeflix.com to get your API keys
stupeflixAccessKey = os.getenv("STUPEFLIX_ACCESS_KEY")
stupeflixSecretKey = os.getenv("STUPEFLIX_SECRET_KEY")
stupeflixHost = os.getenv("STUPEFLIX_HOST") or 'http://services.stupeflix.com/'




#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
r""" A JSON data encoder and decoder.

 This Python module implements the JSON (http://json.org/) data
 encoding format; a subset of ECMAScript (aka JavaScript) for encoding
 primitive data types (numbers, strings, booleans, lists, and
 associative arrays) in a language-neutral simple text-based syntax.
 
 It can encode or decode between JSON formatted strings and native
 Python data types.  Normally you would use the encode() and decode()
 functions defined by this module, but if you want more control over
 the processing you can use the JSON class.

 This implementation tries to be as completely cormforming to all
 intricacies of the standards as possible.  It can operate in strict
 mode (which only allows JSON-compliant syntax) or a non-strict mode
 (which allows much more of the whole ECMAScript permitted syntax).
 This includes complete support for Unicode strings (including
 surrogate-pairs for non-BMP characters), and all number formats
 including negative zero and IEEE 754 non-numbers such a NaN or
 Infinity.

 The JSON/ECMAScript to Python type mappings are:
    ---JSON---             ---Python---
    null                   None
    undefined              undefined  (note 1)
    Boolean (true,false)   bool  (True or False)
    Integer                int or long  (note 2)
    Float                  float
    String                 str or unicode  ( "..." or u"..." )
    Array [a, ...]         list  ( [...] )
    Object {a:b, ...}      dict  ( {...} )
    
    -- Note 1. an 'undefined' object is declared in this module which
       represents the native Python value for this type when in
       non-strict mode.

    -- Note 2. some ECMAScript integers may be up-converted to Python
       floats, such as 1e+40.  Also integer -0 is converted to
       float -0, so as to preserve the sign (which ECMAScript requires).

 In addition, when operating in non-strict mode, several IEEE 754
 non-numbers are also handled, and are mapped to specific Python
 objects declared in this module:

     NaN (not a number)     nan    (float('nan'))
     Infinity, +Infinity    inf    (float('inf'))
     -Infinity              neginf (float('-inf'))

 When encoding Python objects into JSON, you may use types other than
 native lists or dictionaries, as long as they support the minimal
 interfaces required of all sequences or mappings.  This means you can
 use generators and iterators, tuples, UserDict subclasses, etc.

 To make it easier to produce JSON encoded representations of user
 defined classes, if the object has a method named json_equivalent(),
 then it will call that method and attempt to encode the object
 returned from it instead.  It will do this recursively as needed and
 before any attempt to encode the object using it's default
 strategies.  Note that any json_equivalent() method should return
 "equivalent" Python objects to be encoded, not an already-encoded
 JSON-formatted string.  There is no such aid provided to decode
 JSON back into user-defined classes as that would dramatically
 complicate the interface.
 
 When decoding strings with this module it may operate in either
 strict or non-strict mode.  The strict mode only allows syntax which
 is conforming to RFC 4627 (JSON), while the non-strict allows much
 more of the permissible ECMAScript syntax.

 The following are permitted when processing in NON-STRICT mode:

    * Unicode format control characters are allowed anywhere in the input.
    * All Unicode line terminator characters are recognized.
    * All Unicode white space characters are recognized.
    * The 'undefined' keyword is recognized.
    * Hexadecimal number literals are recognized (e.g., 0xA6, 0177).
    * String literals may use either single or double quote marks.
    * Strings may contain \x (hexadecimal) escape sequences, as well as the
      \v and \0 escape sequences.
    * Lists may have omitted (elided) elements, e.g., [,,,,,], with
      missing elements interpreted as 'undefined' values.
    * Object properties (dictionary keys) can be of any of the
      types: string literals, numbers, or identifiers (the later of
      which are treated as if they are string literals)---as permitted
      by ECMAScript.  JSON only permits strings literals as keys.

 Concerning non-strict and non-ECMAScript allowances:

    * Octal numbers: If you allow the 'octal_numbers' behavior (which
      is never enabled by default), then you can use octal integers
      and octal character escape sequences (per the ECMAScript
      standard Annex B.1.2).  This behavior is allowed, if enabled,
      because it was valid JavaScript at one time.

    * Multi-line string literals:  Strings which are more than one
      line long (contain embedded raw newline characters) are never
      permitted. This is neither valid JSON nor ECMAScript.  Some other
      JSON implementations may allow this, but this module considers
      that behavior to be a mistake.

 References:
    * JSON (JavaScript Object Notation)
      <http://json.org/>
    * RFC 4627. The application/json Media Type for JavaScript Object Notation (JSON)
      <http://www.ietf.org/rfc/rfc4627.txt>
    * ECMA-262 3rd edition (1999)
      <http://www.ecma-international.org/publications/files/ecma-st/ECMA-262.pdf>
    * IEEE 754-1985: Standard for Binary Floating-Point Arithmetic.
      <http://www.cs.berkeley.edu/~ejr/Projects/ieee754/>
    
"""

__author__ = "Deron Meranda <http://deron.meranda.us/>"
__date__ = "2008-03-19"
__version__ = "1.3"
__credits__ = """Copyright (c) 2006-2008 Deron E. Meranda <http://deron.meranda.us/>
Licensed under GNU GPL 3.0 or later.  See LICENSE.txt included with this software.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""

# ------------------------------
# useful global constants

content_type = 'application/json'
file_ext = 'json'
hexdigits = '0123456789ABCDEFabcdef'
octaldigits = '01234567'

# ----------------------------------------------------------------------
# Decimal and float types.
#
# If a JSON number can not be stored in a Python float without loosing
# precision and the Python has the decimal type, then we will try to
# use decimal instead of float.  To make this determination we need to
# know the limits of the float type, but Python doesn't have an easy
# way to tell what the largest floating-point number it supports.  So,
# we detemine the precision and scale of the float type by testing it.

try:
    # decimal module was introduced in Python 2.4
    import decimal
except ImportError:
    decimal = None

def determine_float_precision():
    """Returns a tuple (significant_digits, max_exponent) for the float type.
    """
    import math
    # Just count the digits in pi.  The last two decimal digits
    # may only be partial digits, so discount for them.
    whole, frac = repr(math.pi).split('.')
    sigdigits = len(whole) + len(frac) - 2

    # This is a simple binary search.  We find the largest exponent
    # that the float() type can handle without going infinite or
    # raising errors.
    maxexp = None
    minv = 0; maxv = 1000
    while True:
        if minv+1 == maxv:
            maxexp = minv - 1
            break
        elif maxv < minv:
            maxexp = None
            break
        m = (minv + maxv) // 2
        try:
            f = repr(float( '1e+%d' % m ))
        except ValueError:
            f = None
        else:
            if not f or f[0] < '0' or f[0] > '9':
                f = None
        if not f:
            # infinite
            maxv = m
        else:
            minv = m
    return sigdigits, maxexp

float_sigdigits, float_maxexp = determine_float_precision()

# ----------------------------------------------------------------------
# The undefined value.
#
# ECMAScript has an undefined value (similar to yet distinct from null).
# Neither Python or strict JSON have support undefined, but to allow
# JavaScript behavior we must simulate it.

class _undefined_class(object):
    """Represents the ECMAScript 'undefined' value."""
    __slots__ = []
    def __repr__(self):
        return self.__module__ + '.undefined'
    def __str__(self):
        return 'undefined'
    def __nonzero__(self):
        return False
undefined = _undefined_class()
del _undefined_class


# ----------------------------------------------------------------------
# Non-Numbers: NaN, Infinity, -Infinity
#
# ECMAScript has official support for non-number floats, although
# strict JSON does not.  Python doesn't either.  So to support the
# full JavaScript behavior we must try to add them into Python, which
# is unfortunately a bit of black magic.  If our python implementation
# happens to be built on top of IEEE 754 we can probably trick python
# into using real floats.  Otherwise we must simulate it with classes.

def _nonnumber_float_constants():
    """Try to return the Nan, Infinity, and -Infinity float values.
    
    This is unnecessarily complex because there is no standard
    platform- independent way to do this in Python as the language
    (opposed to some implementation of it) doesn't discuss
    non-numbers.  We try various strategies from the best to the
    worst.
    
    If this Python interpreter uses the IEEE 754 floating point
    standard then the returned values will probably be real instances
    of the 'float' type.  Otherwise a custom class object is returned
    which will attempt to simulate the correct behavior as much as
    possible.

    """
    try:
        # First, try (mostly portable) float constructor.  Works under
        # Linux x86 (gcc) and some Unices.
        nan = float('nan')
        inf = float('inf')
        neginf = float('-inf')
    except ValueError:
        try:
            # Try the AIX (PowerPC) float constructors
            nan = float('NaNQ')
            inf = float('INF')
            neginf = float('-INF')
        except ValueError:
            try:
                # Next, try binary unpacking.  Should work under
                # platforms using IEEE 754 floating point.
                import struct, sys
                xnan = '7ff8000000000000'.decode('hex')  # Quiet NaN
                xinf = '7ff0000000000000'.decode('hex')
                xcheck = 'bdc145651592979d'.decode('hex') # -3.14159e-11
                # Could use float.__getformat__, but it is a new python feature,
                # so we use sys.byteorder.
                if sys.byteorder == 'big':
                    nan = struct.unpack('d', xnan)[0]
                    inf = struct.unpack('d', xinf)[0]
                    check = struct.unpack('d', xcheck)[0]
                else:
                    nan = struct.unpack('d', xnan[::-1])[0]
                    inf = struct.unpack('d', xinf[::-1])[0]
                    check = struct.unpack('d', xcheck[::-1])[0]
                neginf = - inf
                if check != -3.14159e-11:
                    raise ValueError('Unpacking raw IEEE 754 floats does not work')
            except (ValueError, TypeError):
                # Punt, make some fake classes to simulate.  These are
                # not perfect though.  For instance nan * 1.0 == nan,
                # as expected, but 1.0 * nan == 0.0, which is wrong.
                class nan(float):
                    """An approximation of the NaN (not a number) floating point number."""
                    def __repr__(self): return 'nan'
                    def __str__(self): return 'nan'
                    def __add__(self,x): return self
                    def __radd__(self,x): return self
                    def __sub__(self,x): return self
                    def __rsub__(self,x): return self
                    def __mul__(self,x): return self
                    def __rmul__(self,x): return self
                    def __div__(self,x): return self
                    def __rdiv__(self,x): return self
                    def __divmod__(self,x): return (self,self)
                    def __rdivmod__(self,x): return (self,self)
                    def __mod__(self,x): return self
                    def __rmod__(self,x): return self
                    def __pow__(self,exp): return self
                    def __rpow__(self,exp): return self
                    def __neg__(self): return self
                    def __pos__(self): return self
                    def __abs__(self): return self
                    def __lt__(self,x): return False
                    def __le__(self,x): return False
                    def __eq__(self,x): return False
                    def __neq__(self,x): return True
                    def __ge__(self,x): return False
                    def __gt__(self,x): return False
                    def __complex__(self,*a): raise NotImplementedError('NaN can not be converted to a complex')
                if decimal:
                    nan = decimal.Decimal('NaN')
                else:
                    nan = nan()
                class inf(float):
                    """An approximation of the +Infinity floating point number."""
                    def __repr__(self): return 'inf'
                    def __str__(self): return 'inf'
                    def __add__(self,x): return self
                    def __radd__(self,x): return self
                    def __sub__(self,x): return self
                    def __rsub__(self,x): return self
                    def __mul__(self,x):
                        if x is neginf or x < 0:
                            return neginf
                        elif x == 0:
                            return nan
                        else:
                            return self
                    def __rmul__(self,x): return self.__mul__(x)
                    def __div__(self,x):
                        if x == 0:
                            raise ZeroDivisionError('float division')
                        elif x < 0:
                            return neginf
                        else:
                            return self
                    def __rdiv__(self,x):
                        if x is inf or x is neginf or x is nan:
                            return nan
                        return 0.0
                    def __divmod__(self,x):
                        if x == 0:
                            raise ZeroDivisionError('float divmod()')
                        elif x < 0:
                            return (nan,nan)
                        else:
                            return (self,self)
                    def __rdivmod__(self,x):
                        if x is inf or x is neginf or x is nan:
                            return (nan, nan)
                        return (0.0, x)
                    def __mod__(self,x):
                        if x == 0:
                            raise ZeroDivisionError('float modulo')
                        else:
                            return nan
                    def __rmod__(self,x):
                        if x is inf or x is neginf or x is nan:
                            return nan
                        return x
                    def __pow__(self, exp):
                        if exp == 0:
                            return 1.0
                        else:
                            return self
                    def __rpow__(self, x):
                        if -1 < x < 1: return 0.0
                        elif x == 1.0: return 1.0
                        elif x is nan or x is neginf or x < 0:
                            return nan
                        else:
                            return self
                    def __neg__(self): return neginf
                    def __pos__(self): return self
                    def __abs__(self): return self
                    def __lt__(self,x): return False
                    def __le__(self,x):
                        if x is self:
                            return True
                        else:
                            return False
                    def __eq__(self,x):
                        if x is self:
                            return True
                        else:
                            return False
                    def __neq__(self,x):
                        if x is self:
                            return False
                        else:
                            return True
                    def __ge__(self,x): return True
                    def __gt__(self,x): return True
                    def __complex__(self,*a): raise NotImplementedError('Infinity can not be converted to a complex')
                if decimal:
                    inf = decimal.Decimal('Infinity')
                else:
                    inf = inf()
                class neginf(float):
                    """An approximation of the -Infinity floating point number."""
                    def __repr__(self): return '-inf'
                    def __str__(self): return '-inf'
                    def __add__(self,x): return self
                    def __radd__(self,x): return self
                    def __sub__(self,x): return self
                    def __rsub__(self,x): return self
                    def __mul__(self,x):
                        if x is self or x < 0:
                            return inf
                        elif x == 0:
                            return nan
                        else:
                            return self
                    def __rmul__(self,x): return self.__mul__(self)
                    def __div__(self,x):
                        if x == 0:
                            raise ZeroDivisionError('float division')
                        elif x < 0:
                            return inf
                        else:
                            return self
                    def __rdiv__(self,x):
                        if x is inf or x is neginf or x is nan:
                            return nan
                        return -0.0
                    def __divmod__(self,x):
                        if x == 0:
                            raise ZeroDivisionError('float divmod()')
                        elif x < 0:
                            return (nan,nan)
                        else:
                            return (self,self)
                    def __rdivmod__(self,x):
                        if x is inf or x is neginf or x is nan:
                            return (nan, nan)
                        return (-0.0, x)
                    def __mod__(self,x):
                        if x == 0:
                            raise ZeroDivisionError('float modulo')
                        else:
                            return nan
                    def __rmod__(self,x):
                        if x is inf or x is neginf or x is nan:
                            return nan
                        return x
                    def __pow__(self,exp):
                        if exp == 0:
                            return 1.0
                        else:
                            return self
                    def __rpow__(self, x):
                        if x is nan or x is inf or x is inf:
                            return nan
                        return 0.0
                    def __neg__(self): return inf
                    def __pos__(self): return self
                    def __abs__(self): return inf
                    def __lt__(self,x): return True
                    def __le__(self,x): return True
                    def __eq__(self,x):
                        if x is self:
                            return True
                        else:
                            return False
                    def __neq__(self,x):
                        if x is self:
                            return False
                        else:
                            return True
                    def __ge__(self,x):
                        if x is self:
                            return True
                        else:
                            return False
                    def __gt__(self,x): return False
                    def __complex__(self,*a): raise NotImplementedError('-Infinity can not be converted to a complex')
                if decimal:
                    neginf = decimal.Decimal('-Infinity')
                else:
                    neginf = neginf(0)
    return nan, inf, neginf

nan, inf, neginf = _nonnumber_float_constants()
del _nonnumber_float_constants


# ----------------------------------------------------------------------
# String processing helpers

unsafe_string_chars = '"\\' + ''.join([chr(i) for i in range(0x20)])
def skipstringsafe( s, start=0, end=None ):
    i = start
    #if end is None:
    #    end = len(s)
    while i < end and s[i] not in unsafe_string_chars:
        #c = s[i]
        #if c in unsafe_string_chars:
        #    break
        i += 1
    return i
def skipstringsafe_slow( s, start=0, end=None ):
    i = start
    if end is None:
        end = len(s)
    while i < end:
        c = s[i]
        if c == '"' or c == '\\' or ord(c) <= 0x1f:
            break
        i += 1
    return i

def extend_list_with_sep( orig_seq, extension_seq, sepchar='' ):
    if not sepchar:
        orig_seq.extend( extension_seq )
    else:
        for i, x in enumerate(extension_seq):
            if i > 0:
                orig_seq.append( sepchar )
            orig_seq.append( x )

def extend_and_flatten_list_with_sep( orig_seq, extension_seq, separator='' ):
    for i, part in enumerate(extension_seq):
        if i > 0 and separator:
            orig_seq.append( separator )
        orig_seq.extend( part )


# ----------------------------------------------------------------------
# Unicode helpers
#
# JSON requires that all JSON implementations must support the UTF-32
# encoding (as well as UTF-8 and UTF-16).  But earlier versions of
# Python did not provide a UTF-32 codec.  So we must implement UTF-32
# ourselves in case we need it.

def utf32le_encode( obj, errors='strict' ):
    """Encodes a Unicode string into a UTF-32LE encoded byte string."""
    import struct
    try:
        import cStringIO as sio
    except ImportError:
        import StringIO as sio
    f = sio.StringIO()
    write = f.write
    pack = struct.pack
    for c in obj:
        n = ord(c)
        if 0xD800 <= n <= 0xDFFF: # surrogate codepoints are prohibited by UTF-32
            if errors == 'ignore':
                continue
            elif errors == 'replace':
                n = ord('?')
            else:
                cname = 'U+%04X'%n
                raise UnicodeError('UTF-32 can not encode surrogate characters',cname)
        write( pack('<L', n) )
    return f.getvalue()


def utf32be_encode( obj, errors='strict' ):
    """Encodes a Unicode string into a UTF-32BE encoded byte string."""
    import struct
    try:
        import cStringIO as sio
    except ImportError:
        import StringIO as sio
    f = sio.StringIO()
    write = f.write
    pack = struct.pack
    for c in obj:
        n = ord(c)
        if 0xD800 <= n <= 0xDFFF: # surrogate codepoints are prohibited by UTF-32
            if errors == 'ignore':
                continue
            elif errors == 'replace':
                n = ord('?')
            else:
                cname = 'U+%04X'%n
                raise UnicodeError('UTF-32 can not encode surrogate characters',cname)
        write( pack('>L', n) )
    return f.getvalue()


def utf32le_decode( obj, errors='strict' ):
    """Decodes a UTF-32LE byte string into a Unicode string."""
    if len(obj) % 4 != 0:
        raise UnicodeError('UTF-32 decode error, data length not a multiple of 4 bytes')
    import struct
    unpack = struct.unpack
    chars = []
    i = 0
    for i in range(0, len(obj), 4):
        seq = obj[i:i+4]
        n = unpack('<L',seq)[0]
        chars.append( unichr(n) )
    return u''.join( chars )


def utf32be_decode( obj, errors='strict' ):
    """Decodes a UTF-32BE byte string into a Unicode string."""
    if len(obj) % 4 != 0:
        raise UnicodeError('UTF-32 decode error, data length not a multiple of 4 bytes')
    import struct
    unpack = struct.unpack
    chars = []
    i = 0
    for i in range(0, len(obj), 4):
        seq = obj[i:i+4]
        n = unpack('>L',seq)[0]
        chars.append( unichr(n) )
    return u''.join( chars )


def auto_unicode_decode( s ):
    """Takes a string and tries to convert it to a Unicode string.

    This will return a Python unicode string type corresponding to the
    input string (either str or unicode).  The character encoding is
    guessed by looking for either a Unicode BOM prefix, or by the
    rules specified by RFC 4627.  When in doubt it is assumed the
    input is encoded in UTF-8 (the default for JSON).

    """
    if isinstance(s, unicode):
        return s
    if len(s) < 4:
        return s.decode('utf8')  # not enough bytes, assume default of utf-8
    # Look for BOM marker
    import codecs
    bom2 = s[:2]
    bom4 = s[:4]
    a, b, c, d = map(ord, s[:4])  # values of first four bytes
    if bom4 == codecs.BOM_UTF32_LE:
        encoding = 'utf-32le'
        s = s[4:]
    elif bom4 == codecs.BOM_UTF32_BE:
        encoding = 'utf-32be'
        s = s[4:]
    elif bom2 == codecs.BOM_UTF16_LE:
        encoding = 'utf-16le'
        s = s[2:]
    elif bom2 == codecs.BOM_UTF16_BE:
        encoding = 'utf-16be'
        s = s[2:]
    # No BOM, so autodetect encoding used by looking at first four bytes
    # according to RFC 4627 section 3.
    elif a==0 and b==0 and c==0 and d!=0: # UTF-32BE
        encoding = 'utf-32be'
    elif a==0 and b!=0 and c==0 and d!=0: # UTF-16BE
        encoding = 'utf-16be'
    elif a!=0 and b==0 and c==0 and d==0: # UTF-32LE
        encoding = 'utf-32le'
    elif a!=0 and b==0 and c!=0 and d==0: # UTF-16LE
        encoding = 'utf-16le'
    else: #if a!=0 and b!=0 and c!=0 and d!=0: # UTF-8
        # JSON spec says default is UTF-8, so always guess it
        # if we can't guess otherwise
        encoding = 'utf8'
    # Make sure the encoding is supported by Python
    try:
        cdk = codecs.lookup(encoding)
    except LookupError:
        if encoding.startswith('utf-32') \
               or encoding.startswith('ucs4') \
               or encoding.startswith('ucs-4'):
            # Python doesn't natively have a UTF-32 codec, but JSON
            # requires that it be supported.  So we must decode these
            # manually.
            if encoding.endswith('le'):
                unis = utf32le_decode(s)
            else:
                unis = utf32be_decode(s)
        else:
            raise JSONDecodeError('this python has no codec for this character encoding',encoding)
    else:
        # Convert to unicode using a standard codec
        unis = s.decode(encoding)
    return unis


def surrogate_pair_as_unicode( c1, c2 ):
    """Takes a pair of unicode surrogates and returns the equivalent unicode character.

    The input pair must be a surrogate pair, with c1 in the range
    U+D800 to U+DBFF and c2 in the range U+DC00 to U+DFFF.

    """
    n1, n2 = ord(c1), ord(c2)
    if n1 < 0xD800 or n1 > 0xDBFF or n2 < 0xDC00 or n2 > 0xDFFF:
        raise JSONDecodeError('illegal Unicode surrogate pair',(c1,c2))
    a = n1 - 0xD800
    b = n2 - 0xDC00
    v = (a << 10) | b
    v += 0x10000
    return unichr(v)


def unicode_as_surrogate_pair( c ):
    """Takes a single unicode character and returns a sequence of surrogate pairs.

    The output of this function is a tuple consisting of one or two unicode
    characters, such that if the input character is outside the BMP range
    then the output is a two-character surrogate pair representing that character.

    If the input character is inside the BMP then the output tuple will have
    just a single character...the same one.

    """
    n = ord(c)
    if n < 0x10000:
        return (unichr(n),)  # in BMP, surrogate pair not required
    v = n - 0x10000
    vh = (v >> 10) & 0x3ff   # highest 10 bits
    vl = v & 0x3ff  # lowest 10 bits
    w1 = 0xD800 | vh
    w2 = 0xDC00 | vl
    return (unichr(w1), unichr(w2))


# ----------------------------------------------------------------------
# Type identification

def isnumbertype( obj ):
    """Is the object of a Python number type (excluding complex)?"""
    return isinstance(obj, (int,long,float)) \
           and not isinstance(obj, bool) \
           or obj is nan or obj is inf or obj is neginf


def isstringtype( obj ):
    """Is the object of a Python string type?"""
    if isinstance(obj, basestring):
        return True
    # Must also check for some other pseudo-string types
    import types, UserString
    return isinstance(obj, types.StringTypes) \
           or isinstance(obj, UserString.UserString) \
           or isinstance(obj, UserString.MutableString)


# ----------------------------------------------------------------------
# Numeric helpers

def decode_hex( hexstring ):
    """Decodes a hexadecimal string into it's integer value."""
    # We don't use the builtin 'hex' codec in python since it can
    # not handle odd numbers of digits, nor raise the same type
    # of exceptions we want to.
    n = 0
    for c in hexstring:
        if '0' <= c <= '9':
            d = ord(c) - ord('0')
        elif 'a' <= c <= 'f':
            d = ord(c) - ord('a') + 10
        elif 'A' <= c <= 'F':
            d = ord(c) - ord('A') + 10
        else:
            raise JSONDecodeError('not a hexadecimal number',hexstring)
        # Could use ((n << 4 ) | d), but python 2.3 issues a FutureWarning.
        n = (n * 16) + d
    return n


def decode_octal( octalstring ):
    """Decodes an octal string into it's integer value."""
    n = 0
    for c in octalstring:
        if '0' <= c <= '7':
            d = ord(c) - ord('0')
        else:
            raise JSONDecodeError('not an octal number',octalstring)
        # Could use ((n << 3 ) | d), but python 2.3 issues a FutureWarning.
        n = (n * 8) + d
    return n


# ----------------------------------------------------------------------
# Exception classes.

class JSONError(ValueError):
    """Our base class for all JSON-related errors.

    """
    def pretty_description(self):
        err = self.args[0]
        if len(self.args) > 1:
            err += ': '
            for anum, a in enumerate(self.args[1:]):
                if anum > 1:
                    err += ', '
                astr = repr(a)
                if len(astr) > 20:
                    astr = astr[:20] + '...'
                err += astr
        return err

class JSONDecodeError(JSONError):
    """An exception class raised when a JSON decoding error (syntax error) occurs."""


class JSONEncodeError(JSONError):
    """An exception class raised when a python object can not be encoded as a JSON string."""


#----------------------------------------------------------------------
# The main JSON encoder/decoder class.

class JSON(object):
    """An encoder/decoder for JSON data streams.

    Usually you will call the encode() or decode() methods.  The other
    methods are for lower-level processing.

    Whether the JSON parser runs in strict mode (which enforces exact
    compliance with the JSON spec) or the more forgiving non-string mode
    can be affected by setting the 'strict' argument in the object's
    initialization; or by assigning True or False to the 'strict'
    property of the object.

    You can also adjust a finer-grained control over strictness by
    allowing or preventing specific behaviors.  You can get a list of
    all the available behaviors by accessing the 'behaviors' property.
    Likewise the allowed_behaviors and prevented_behaviors list which
    behaviors will be allowed and which will not.  Call the allow()
    or prevent() methods to adjust these.
    
    """
    _escapes_json = { # character escapes in JSON
        '"': '"',
        '/': '/',
        '\\': '\\',
        'b': '\b',
        'f': '\f',
        'n': '\n',
        'r': '\r',
        't': '\t',
        }

    _escapes_js = { # character escapes in Javascript
        '"': '"',
        '\'': '\'',
        '\\': '\\',
        'b': '\b',
        'f': '\f',
        'n': '\n',
        'r': '\r',
        't': '\t',
        'v': '\v',
        '0': '\x00'
        }

    # Following is a reverse mapping of escape characters, used when we
    # output JSON.  Only those escapes which are always safe (e.g., in JSON)
    # are here.  It won't hurt if we leave questionable ones out.
    _rev_escapes = {'\n': '\\n',
                    '\t': '\\t',
                    '\b': '\\b',
                    '\r': '\\r',
                    '\f': '\\f',
                    '"': '\\"',
                    '\\': '\\\\'}

    def __init__(self, strict=False, compactly=True, escape_unicode=False):
        """Creates a JSON encoder/decoder object.
        
        If 'strict' is set to True, then only strictly-conforming JSON
        output will be produced.  Note that this means that some types
        of values may not be convertable and will result in a
        JSONEncodeError exception.
        
        If 'compactly' is set to True, then the resulting string will
        have all extraneous white space removed; if False then the
        string will be "pretty printed" with whitespace and indentation
        added to make it more readable.
        
        If 'escape_unicode' is set to True, then all non-ASCII characters
        will be represented as a unicode escape sequence; if False then
        the actual real unicode character will be inserted if possible.

        The 'escape_unicode' can also be a function, which when called
        with a single argument of a unicode character will return True
        if the character should be escaped or False if it should not.
        
        If you wish to extend the encoding to ba able to handle
        additional types, you should subclass this class and override
        the encode_default() method.
        
        """
        import sys
        self._set_strictness(strict)
        self._encode_compactly = compactly
        try:
            # see if we were passed a predicate function
            b = escape_unicode(u'A')
            self._encode_unicode_as_escapes = escape_unicode
        except (ValueError, NameError, TypeError):
            # Just set to True or False.  We could use lambda x:True
            # to make it more consistent (always a function), but it
            # will be too slow, so we'll make explicit tests later.
            self._encode_unicode_as_escapes = bool(escape_unicode)
        self._sort_dictionary_keys = True

        # The following is a boolean map of the first 256 characters
        # which will quickly tell us which of those characters never
        # need to be escaped.

        self._asciiencodable = [32 <= c < 128 and not self._rev_escapes.has_key(chr(c))
                              for c in range(0,255)]

    def _set_strictness(self, strict):
        """Changes the strictness behavior.

        Pass True to be very strict about JSON syntax, or False to be looser.
        """
        self._allow_any_type_at_start = not strict
        self._allow_all_numeric_signs = not strict
        self._allow_comments = not strict
        self._allow_control_char_in_string = not strict
        self._allow_hex_numbers = not strict
        self._allow_initial_decimal_point = not strict
        self._allow_js_string_escapes = not strict
        self._allow_non_numbers = not strict
        self._allow_nonescape_characters = not strict  # "\z" -> "z"
        self._allow_nonstring_keys = not strict
        self._allow_omitted_array_elements = not strict
        self._allow_single_quoted_strings = not strict
        self._allow_trailing_comma_in_literal = not strict
        self._allow_undefined_values = not strict
        self._allow_unicode_format_control_chars = not strict
        self._allow_unicode_whitespace = not strict
        # Always disable this by default
        self._allow_octal_numbers = False

    def allow(self, behavior):
        """Allow the specified behavior (turn off a strictness check).

        The list of all possible behaviors is available in the behaviors property.
        You can see which behaviors are currently allowed by accessing the
        allowed_behaviors property.

        """
        p = '_allow_' + behavior
        if hasattr(self, p):
            setattr(self, p, True)
        else:
            raise AttributeError('Behavior is not known',behavior)

    def prevent(self, behavior):
        """Prevent the specified behavior (turn on a strictness check).

        The list of all possible behaviors is available in the behaviors property.
        You can see which behaviors are currently prevented by accessing the
        prevented_behaviors property.

        """
        p = '_allow_' + behavior
        if hasattr(self, p):
            setattr(self, p, False)
        else:
            raise AttributeError('Behavior is not known',behavior)

    def _get_behaviors(self):
        return sorted([ n[len('_allow_'):] for n in self.__dict__ \
                        if n.startswith('_allow_')])
    behaviors = property(_get_behaviors,
                         doc='List of known behaviors that can be passed to allow() or prevent() methods')

    def _get_allowed_behaviors(self):
        return sorted([ n[len('_allow_'):] for n in self.__dict__ \
                        if n.startswith('_allow_') and getattr(self,n)])
    allowed_behaviors = property(_get_allowed_behaviors,
                                 doc='List of known behaviors that are currently allowed')

    def _get_prevented_behaviors(self):
        return sorted([ n[len('_allow_'):] for n in self.__dict__ \
                        if n.startswith('_allow_') and not getattr(self,n)])
    prevented_behaviors = property(_get_prevented_behaviors,
                                   doc='List of known behaviors that are currently prevented')

    def _is_strict(self):
        return not self.allowed_behaviors
    strict = property(_is_strict, _set_strictness,
                      doc='True if adherence to RFC 4627 syntax is strict, or False is more generous ECMAScript syntax is permitted')


    def isws(self, c):
        """Determines if the given character is considered as white space.
        
        Note that Javscript is much more permissive on what it considers
        to be whitespace than does JSON.
        
        Ref. ECMAScript section 7.2

        """
        if not self._allow_unicode_whitespace:
            return c in ' \t\n\r'
        else:
            if not isinstance(c,unicode):
                c = unicode(c)
            if c in u' \t\n\r\f\v':
                return True
            import unicodedata
            return unicodedata.category(c) == 'Zs'

    def islineterm(self, c):
        """Determines if the given character is considered a line terminator.

        Ref. ECMAScript section 7.3

        """
        if c == '\r' or c == '\n':
            return True
        if c == u'\u2028' or c == u'\u2029': # unicodedata.category(c) in  ['Zl', 'Zp']
            return True
        return False

    def strip_format_control_chars(self, txt):
        """Filters out all Unicode format control characters from the string.

        ECMAScript permits any Unicode "format control characters" to
        appear at any place in the source code.  They are to be
        ignored as if they are not there before any other lexical
        tokenization occurs.  Note that JSON does not allow them.

        Ref. ECMAScript section 7.1.

        """
        import unicodedata
        txt2 = filter( lambda c: unicodedata.category(unicode(c)) != 'Cf',
                       txt )
        return txt2


    def decode_null(self, s, i=0):
        """Intermediate-level decoder for ECMAScript 'null' keyword.

        Takes a string and a starting index, and returns a Python
        None object and the index of the next unparsed character.

        """
        if i < len(s) and s[i:i+4] == 'null':
            return None, i+4
        raise JSONDecodeError('literal is not the JSON "null" keyword', s)

    def encode_undefined(self):
        """Produces the ECMAScript 'undefined' keyword."""
        return 'undefined'

    def encode_null(self):
        """Produces the JSON 'null' keyword."""
        return 'null'

    def decode_boolean(self, s, i=0):
        """Intermediate-level decode for JSON boolean literals.

        Takes a string and a starting index, and returns a Python bool
        (True or False) and the index of the next unparsed character.

        """
        if s[i:i+4] == 'true':
            return True, i+4
        elif s[i:i+5] == 'false':
            return False, i+5
        raise JSONDecodeError('literal value is not a JSON boolean keyword',s)

    def encode_boolean(self, b):
        """Encodes the Python boolean into a JSON Boolean literal."""
        if bool(b):
            return 'true'
        return 'false'

    def decode_number(self, s, i=0, imax=None):
        """Intermediate-level decoder for JSON numeric literals.

        Takes a string and a starting index, and returns a Python
        suitable numeric type and the index of the next unparsed character.

        The returned numeric type can be either of a Python int,
        long, or float.  In addition some special non-numbers may
        also be returned such as nan, inf, and neginf (technically
        which are Python floats, but have no numeric value.)

        Ref. ECMAScript section 8.5.

        """
        if imax is None:
            imax = len(s)
        # Detect initial sign character(s)
        if not self._allow_all_numeric_signs:
            if s[i] == '+' or (s[i] == '-' and i+1 < imax and \
                               s[i+1] in '+-'):
                raise JSONDecodeError('numbers in strict JSON may only have a single "-" as a sign prefix',s[i:])
        sign = +1
        j = i  # j will point after the sign prefix
        while j < imax and s[j] in '+-':
            if s[j] == '-': sign = sign * -1
            j += 1
        # Check for ECMAScript symbolic non-numbers
        if s[j:j+3] == 'NaN':
            if self._allow_non_numbers:
                return nan, j+3
            else:
                raise JSONDecodeError('NaN literals are not allowed in strict JSON')
        elif s[j:j+8] == 'Infinity':
            if self._allow_non_numbers:
                if sign < 0:
                    return neginf, j+8
                else:
                    return inf, j+8
            else:
                raise JSONDecodeError('Infinity literals are not allowed in strict JSON')
        elif s[j:j+2] in ('0x','0X'):
            if self._allow_hex_numbers:
                k = j+2
                while k < imax and s[k] in hexdigits:
                    k += 1
                n = sign * decode_hex( s[j+2:k] )
                return n, k
            else:
                raise JSONDecodeError('hexadecimal literals are not allowed in strict JSON',s[i:])
        else:
            # Decimal (or octal) number, find end of number.
            # General syntax is:  \d+[\.\d+][e[+-]?\d+]
            k = j   # will point to end of digit sequence
            could_be_octal = ( k+1 < imax and s[k] == '0' )  # first digit is 0
            decpt = None  # index into number of the decimal point, if any
            ept = None # index into number of the e|E exponent start, if any
            esign = '+' # sign of exponent
            sigdigits = 0 # number of significant digits (approx, counts end zeros)
            while k < imax and (s[k].isdigit() or s[k] in '.+-eE'):
                c = s[k]
                if c not in octaldigits:
                    could_be_octal = False
                if c == '.':
                    if decpt is not None or ept is not None:
                        break
                    else:
                        decpt = k-j
                elif c in 'eE':
                    if ept is not None:
                        break
                    else:
                        ept = k-j
                elif c in '+-':
                    if not ept:
                        break
                    esign = c
                else: #digit
                    if not ept:
                        sigdigits += 1
                k += 1
            number = s[j:k]  # The entire number as a string
            #print 'NUMBER IS: ', repr(number), ', sign', sign, ', esign', esign, \
            #      ', sigdigits', sigdigits, \
            #      ', decpt', decpt, ', ept', ept

            # Handle octal integers first as an exception.  If octal
            # is not enabled (the ECMAScipt standard) then just do
            # nothing and treat the string as a decimal number.
            if could_be_octal and self._allow_octal_numbers:
                n = sign * decode_octal( number )
                return n, k

            # A decimal number.  Do a quick check on JSON syntax restrictions.
            if number[0] == '.' and not self._allow_initial_decimal_point:
                raise JSONDecodeError('numbers in strict JSON must have at least one digit before the decimal point',s[i:])
            elif number[0] == '0' and \
                     len(number) > 1 and number[1].isdigit():
                if self._allow_octal_numbers:
                    raise JSONDecodeError('initial zero digit is only allowed for octal integers',s[i:])
                else:
                    raise JSONDecodeError('initial zero digit must not be followed by other digits (octal numbers are not permitted)',s[i:])
            # Make sure decimal point is followed by a digit
            if decpt is not None:
                if decpt+1 >= len(number) or not number[decpt+1].isdigit():
                    raise JSONDecodeError('decimal point must be followed by at least one digit',s[i:])
            # Determine the exponential part
            if ept is not None:
                if ept+1 >= len(number):
                    raise JSONDecodeError('exponent in number is truncated',s[i:])
                try:
                    exponent = int(number[ept+1:])
                except ValueError:
                    raise JSONDecodeError('not a valid exponent in number',s[i:])
                ##print 'EXPONENT', exponent
            else:
                exponent = 0
            # Try to make an int/long first.
            if decpt is None and exponent >= 0:
                # An integer
                if ept:
                    n = int(number[:ept])
                else:
                    n = int(number)
                n *= sign
                if exponent:
                    n *= 10**exponent
                if n == 0 and sign < 0:
                    # minus zero, must preserve negative sign so make a float
                    n = -0.0
            else:
                try:
                    if decimal and (abs(exponent) > float_maxexp or sigdigits > float_sigdigits):
                        try:
                            n = decimal.Decimal(number)
                            n = n.normalize()
                        except decimal.Overflow:
                            if sign<0:
                                n = neginf
                            else:
                                n = inf
                        else:
                            n *= sign
                    else:
                        n = float(number) * sign
                except ValueError:
                    raise JSONDecodeError('not a valid JSON numeric literal', s[i:j])
            return n, k

    def encode_number(self, n):
        """Encodes a Python numeric type into a JSON numeric literal.
        
        The special non-numeric values of float('nan'), float('inf')
        and float('-inf') are translated into appropriate JSON
        literals.
        
        Note that Python complex types are not handled, as there is no
        ECMAScript equivalent type.
        
        """
        if isinstance(n, complex):
            if n.imag:
                raise JSONEncodeError('Can not encode a complex number that has a non-zero imaginary part',n)
            n = n.real
        if isinstance(n, (int,long)):
            return str(n)
        if decimal and isinstance(n, decimal.Decimal):
            return str(n)
        global nan, inf, neginf
        if n is nan:
            return 'NaN'
        elif n is inf:
            return 'Infinity'
        elif n is neginf:
            return '-Infinity'
        elif isinstance(n, float):
            # Check for non-numbers.
            # In python nan == inf == -inf, so must use repr() to distinguish
            reprn = repr(n).lower()
            if ('inf' in reprn and '-' in reprn) or n == neginf:
                return '-Infinity'
            elif 'inf' in reprn or n is inf:
                return 'Infinity'
            elif 'nan' in reprn or n is nan:
                return 'NaN'
            return repr(n)
        else:
            raise TypeError('encode_number expected an integral, float, or decimal number type',type(n))

    def decode_string(self, s, i=0, imax=None):
        """Intermediate-level decoder for JSON string literals.

        Takes a string and a starting index, and returns a Python
        string (or unicode string) and the index of the next unparsed
        character.

        """
        if imax is None:
            imax = len(s)
        if imax < i+2 or s[i] not in '"\'':
            raise JSONDecodeError('string literal must be properly quoted',s[i:])
        closer = s[i]
        if closer == '\'' and not self._allow_single_quoted_strings:
            raise JSONDecodeError('string literals must use double quotation marks in strict JSON',s[i:])
        i += 1 # skip quote
        if self._allow_js_string_escapes:
            escapes = self._escapes_js
        else:
            escapes = self._escapes_json
        ccallowed = self._allow_control_char_in_string
        chunks = []
        _append = chunks.append
        done = False
        high_surrogate = None
        while i < imax:
            c = s[i]
            # Make sure a high surrogate is immediately followed by a low surrogate
            if high_surrogate and (i+1 >= imax or s[i:i+2] != '\\u'):
                raise JSONDecodeError('High unicode surrogate must be followed by a low surrogate',s[i:])
            if c == closer:
                i += 1 # skip end quote
                done = True
                break
            elif c == '\\':
                # Escaped character
                i += 1
                if i >= imax:
                    raise JSONDecodeError('escape in string literal is incomplete',s[i-1:])
                c = s[i]

                if '0' <= c <= '7' and self._allow_octal_numbers:
                    # Handle octal escape codes first so special \0 doesn't kick in yet.
                    # Follow Annex B.1.2 of ECMAScript standard.
                    if '0' <= c <= '3':
                        maxdigits = 3
                    else:
                        maxdigits = 2
                    for k in range(i, i+maxdigits+1):
                        if k >= imax or s[k] not in octaldigits:
                            break
                    n = decode_octal(s[i:k])
                    if n < 128:
                        _append( chr(n) )
                    else:
                        _append( unichr(n) )
                    i = k
                    continue

                if escapes.has_key(c):
                    _append(escapes[c])
                    i += 1
                elif c == 'u' or c == 'x':
                    i += 1
                    if c == 'u':
                        digits = 4
                    else: # c== 'x'
                        if not self._allow_js_string_escapes:
                            raise JSONDecodeError(r'string literals may not use the \x hex-escape in strict JSON',s[i-1:])
                        digits = 2
                    if i+digits >= imax:
                        raise JSONDecodeError('numeric character escape sequence is truncated',s[i-1:])
                    n = decode_hex( s[i:i+digits] )
                    if high_surrogate:
                        # Decode surrogate pair and clear high surrogate
                        _append( surrogate_pair_as_unicode( high_surrogate, unichr(n) ) )
                        high_surrogate = None
                    elif n < 128:
                        # ASCII chars always go in as a str
                        _append( chr(n) )
                    elif 0xd800 <= n <= 0xdbff: # high surrogate
                        if imax < i + digits + 2 or s[i+digits] != '\\' or s[i+digits+1] != 'u':
                            raise JSONDecodeError('High unicode surrogate must be followed by a low surrogate',s[i-2:])
                        high_surrogate = unichr(n)  # remember until we get to the low surrogate
                    elif 0xdc00 <= n <= 0xdfff: # low surrogate
                        raise JSONDecodeError('Low unicode surrogate must be proceeded by a high surrogate',s[i-2:])
                    else:
                        # Other chars go in as a unicode char
                        _append( unichr(n) )
                    i += digits
                else:
                    # Unknown escape sequence
                    if self._allow_nonescape_characters:
                        _append( c )
                        i += 1
                    else:
                        raise JSONDecodeError('unsupported escape code in JSON string literal',s[i-1:])
            elif ord(c) <= 0x1f: # A control character
                if self.islineterm(c):
                    raise JSONDecodeError('line terminator characters must be escaped inside string literals',s[i:])
                elif ccallowed:
                    _append( c )
                    i += 1
                else:
                    raise JSONDecodeError('control characters must be escaped inside JSON string literals',s[i:])
            else: # A normal character; not an escape sequence or end-quote.
                # Find a whole sequence of "safe" characters so we can append them
                # all at once rather than one a time, for speed.
                j = i
                i += 1
                while i < imax and s[i] not in unsafe_string_chars and s[i] != closer:
                    i += 1
                _append(s[j:i])
        if not done:
            raise JSONDecodeError('string literal is not terminated with a quotation mark',s)
        s = ''.join( chunks )
        return s, i

    def encode_string(self, s):
        """Encodes a Python string into a JSON string literal.

        """
        # Must handle instances of UserString specially in order to be
        # able to use ord() on it's simulated "characters".
        import UserString
        if isinstance(s, (UserString.UserString, UserString.MutableString)):
            def tochar(c):
                return c.data
        else:
            # Could use "lambda c:c", but that is too slow.  So we set to None
            # and use an explicit if test inside the loop.
            tochar = None
        
        chunks = []
        chunks.append('"')
        revesc = self._rev_escapes
        asciiencodable = self._asciiencodable
        encunicode = self._encode_unicode_as_escapes
        i = 0
        imax = len(s)
        while i < imax:
            if tochar:
                c = tochar(s[i])
            else:
                c = s[i]
            cord = ord(c)
            if cord < 256 and asciiencodable[cord] and isinstance(encunicode, bool):
                # Contiguous runs of plain old printable ASCII can be copied
                # directly to the JSON output without worry (unless the user
                # has supplied a custom is-encodable function).
                j = i
                i += 1
                while i < imax:
                    if tochar:
                        c = tochar(s[i])
                    else:
                        c = s[i]
                    cord = ord(c)
                    if cord < 256 and asciiencodable[cord]:
                        i += 1
                    else:
                        break
                chunks.append( unicode(s[j:i]) )
            elif revesc.has_key(c):
                # Has a shortcut escape sequence, like "\n"
                chunks.append(revesc[c])
                i += 1
            elif cord <= 0x1F:
                # Always unicode escape ASCII-control characters
                chunks.append(r'\u%04x' % cord)
                i += 1
            elif 0xD800 <= cord <= 0xDFFF:
                # A raw surrogate character!  This should never happen
                # and there's no way to include it in the JSON output.
                # So all we can do is complain.
                cname = 'U+%04X' % cord
                raise JSONEncodeError('can not include or escape a Unicode surrogate character',cname)
            elif cord <= 0xFFFF:
                # Other BMP Unicode character
                if isinstance(encunicode, bool):
                    doesc = encunicode
                else:
                    doesc = encunicode( c )
                if doesc:
                    chunks.append(r'\u%04x' % cord)
                else:
                    chunks.append( c )
                i += 1
            else: # ord(c) >= 0x10000
                # Non-BMP Unicode
                if isinstance(encunicode, bool):
                    doesc = encunicode
                else:
                    doesc = encunicode( c )
                if doesc:
                    for surrogate in unicode_as_surrogate_pair(c):
                        chunks.append(r'\u%04x' % ord(surrogate))
                else:
                    chunks.append( c )
                i += 1
        chunks.append('"')
        return ''.join( chunks )

    def skip_comment(self, txt, i=0):
        """Skips an ECMAScript comment, either // or /* style.

        The contents of the comment are returned as a string, as well
        as the index of the character immediately after the comment.

        """
        if i+1 >= len(txt) or txt[i] != '/' or txt[i+1] not in '/*':
            return None, i
        if not self._allow_comments:
            raise JSONDecodeError('comments are not allowed in strict JSON',txt[i:])
        multiline = (txt[i+1] == '*')
        istart = i
        i += 2
        while i < len(txt):
            if multiline:
                if txt[i] == '*' and i+1 < len(txt) and txt[i+1] == '/':
                    j = i+2
                    break
                elif txt[i] == '/' and i+1 < len(txt) and txt[i+1] == '*':
                    raise JSONDecodeError('multiline /* */ comments may not nest',txt[istart:i+1])
            else:
                if self.islineterm(txt[i]):
                    j = i  # line terminator is not part of comment
                    break
            i += 1

        if i >= len(txt):
            if not multiline:
                j = len(txt)  # // comment terminated by end of file is okay
            else:
                raise JSONDecodeError('comment was never terminated',txt[istart:])
        return txt[istart:j], j

    def skipws(self, txt, i=0, imax=None, skip_comments=True):
        """Skips whitespace.
        """
        if not self._allow_comments and not self._allow_unicode_whitespace:
            if imax is None:
                imax = len(txt)
            while i < imax and txt[i] in ' \r\n\t':
                i += 1
            return i
        else:
            return self.skipws_any(txt, i, imax, skip_comments)

    def skipws_any(self, txt, i=0, imax=None, skip_comments=True):
        """Skips all whitespace, including comments and unicode whitespace

        Takes a string and a starting index, and returns the index of the
        next non-whitespace character.

        If skip_comments is True and not running in strict JSON mode, then
        comments will be skipped over just like whitespace.

        """
        if imax is None:
            imax = len(txt)
        while i < imax:
            if txt[i] == '/':
                cmt, i = self.skip_comment(txt, i)
            if i < imax and self.isws(txt[i]):
                i += 1
            else:
                break
        return i

    def decode_composite(self, txt, i=0, imax=None):
        """Intermediate-level JSON decoder for composite literal types (array and object).

        Takes text and a starting index, and returns either a Python list or
        dictionary and the index of the next unparsed character.

        """
        if imax is None:
            imax = len(txt)
        i = self.skipws(txt, i, imax)
        starti = i
        if i >= imax or txt[i] not in '{[':
            raise JSONDecodeError('composite object must start with "[" or "{"',txt[i:])
        if txt[i] == '[':
            isdict = False
            closer = ']'
            obj = []
        else:
            isdict = True
            closer = '}'
            obj = {}
        i += 1 # skip opener
        i = self.skipws(txt, i, imax)

        if i < imax and txt[i] == closer:
            # empty composite
            i += 1
            done = True
        else:
            saw_value = False   # set to false at beginning and after commas
            done = False
            while i < imax:
                i = self.skipws(txt, i, imax)
                if i < imax and (txt[i] == ',' or txt[i] == closer):
                    c = txt[i]
                    i += 1
                    if c == ',':
                        if not saw_value:
                            # no preceeding value, an elided (omitted) element
                            if isdict:
                                raise JSONDecodeError('can not omit elements of an object (dictionary)')
                            if self._allow_omitted_array_elements:
                                if self._allow_undefined_values:
                                    obj.append( undefined )
                                else:
                                    obj.append( None )
                            else:
                                raise JSONDecodeError('strict JSON does not permit omitted array (list) elements',txt[i:])
                        saw_value = False
                        continue
                    else: # c == closer
                        if not saw_value and not self._allow_trailing_comma_in_literal:
                            if isdict:
                                raise JSONDecodeError('strict JSON does not allow a final comma in an object (dictionary) literal',txt[i-2:])
                            else:
                                raise JSONDecodeError('strict JSON does not allow a final comma in an array (list) literal',txt[i-2:])
                        done = True
                        break

                # Decode the item
                if isdict and self._allow_nonstring_keys:
                    r = self.decodeobj(txt, i, identifier_as_string=True)
                else:
                    r = self.decodeobj(txt, i, identifier_as_string=False)
                if r:
                    if saw_value:
                        # two values without a separating comma
                        raise JSONDecodeError('values must be separated by a comma', txt[i:r[1]])
                    saw_value = True
                    i = self.skipws(txt, r[1], imax)
                    if isdict:
                        key = r[0]  # Ref 11.1.5
                        if not isstringtype(key):
                            if isnumbertype(key):
                                if not self._allow_nonstring_keys:
                                    raise JSONDecodeError('strict JSON only permits string literals as object properties (dictionary keys)',txt[starti:])
                            else:
                                raise JSONDecodeError('object properties (dictionary keys) must be either string literals or numbers',txt[starti:])
                        if i >= imax or txt[i] != ':':
                            raise JSONDecodeError('object property (dictionary key) has no value, expected ":"',txt[starti:])
                        i += 1
                        i = self.skipws(txt, i, imax)
                        rval = self.decodeobj(txt, i)
                        if rval:
                            i = self.skipws(txt, rval[1], imax)
                            obj[key] = rval[0]
                        else:
                            raise JSONDecodeError('object property (dictionary key) has no value',txt[starti:])
                    else: # list
                        obj.append( r[0] )
                else: # not r
                    if isdict:
                        raise JSONDecodeError('expected a value, or "}"',txt[i:])
                    elif not self._allow_omitted_array_elements:
                        raise JSONDecodeError('expected a value or "]"',txt[i:])
                    else:
                        raise JSONDecodeError('expected a value, "," or "]"',txt[i:])
            # end while
        if not done:
            if isdict:
                raise JSONDecodeError('object literal (dictionary) is not terminated',txt[starti:])
            else:
                raise JSONDecodeError('array literal (list) is not terminated',txt[starti:])
        return obj, i

    def decode_javascript_identifier(self, name):
        """Convert a JavaScript identifier into a Python string object.

        This method can be overriden by a subclass to redefine how JavaScript
        identifiers are turned into Python objects.  By default this just
        converts them into strings.

        """
        return name

    def decodeobj(self, txt, i=0, imax=None, identifier_as_string=False, only_object_or_array=False):
        """Intermediate-level JSON decoder.

        Takes a string and a starting index, and returns a two-tuple consting
        of a Python object and the index of the next unparsed character.

        If there is no value at all (empty string, etc), the None is
        returned instead of a tuple.

        """
        if imax is None:
            imax = len(txt)
        obj = None
        i = self.skipws(txt, i, imax)
        if i >= imax:
            raise JSONDecodeError('Unexpected end of input')
        c = txt[i]

        if c == '[' or c == '{':
            obj, i = self.decode_composite(txt, i, imax)
        elif only_object_or_array:
            raise JSONDecodeError('JSON document must start with an object or array type only', txt[i:i+20])
        elif c == '"' or c == '\'':
            obj, i = self.decode_string(txt, i, imax)
        elif c.isdigit() or c in '.+-':
            obj, i = self.decode_number(txt, i, imax)
        elif c.isalpha() or c in'_$':
            j = i
            while j < imax and (txt[j].isalnum() or txt[j] in '_$'):
                j += 1
            kw = txt[i:j]
            if kw == 'null':
                obj, i = None, j
            elif kw == 'true':
                obj, i = True, j
            elif kw == 'false':
                obj, i = False, j
            elif kw == 'undefined':
                if self._allow_undefined_values:
                    obj, i = undefined, j
                else:
                    raise JSONDecodeError('strict JSON does not allow undefined elements',txt[i:])
            elif kw == 'NaN' or kw == 'Infinity':
                obj, i = self.decode_number(txt, i)
            else:
                if identifier_as_string:
                    obj, i = self.decode_javascript_identifier(kw), j
                else:
                    raise JSONDecodeError('unknown keyword or identifier',kw)
        else:
            raise JSONDecodeError('can not decode value',txt[i:])
        return obj, i



    def decode(self, txt):
        """Decodes a JSON-endoded string into a Python object."""
        if self._allow_unicode_format_control_chars:
            txt = self.strip_format_control_chars(txt)
        r = self.decodeobj(txt, 0, only_object_or_array=not self._allow_any_type_at_start)
        if not r:
            raise JSONDecodeError('can not decode value',txt)
        else:
            obj, i = r
            i = self.skipws(txt, i)
            if i < len(txt):
                raise JSONDecodeError('unexpected or extra text',txt[i:])
        return obj

    def encode(self, obj, nest_level=0):
        """Encodes the Python object into a JSON string representation.

        This method will first attempt to encode an object by seeing
        if it has a json_equivalent() method.  If so than it will
        call that method and then recursively attempt to encode
        the object resulting from that call.

        Next it will attempt to determine if the object is a native
        type or acts like a squence or dictionary.  If so it will
        encode that object directly.

        Finally, if no other strategy for encoding the object of that
        type exists, it will call the encode_default() method.  That
        method currently raises an error, but it could be overridden
        by subclasses to provide a hook for extending the types which
        can be encoded.

        """
        chunks = []
        self.encode_helper(chunks, obj, nest_level)
        return ''.join( chunks )

    def encode_helper(self, chunklist, obj, nest_level):
        #print 'encode_helper(chunklist=%r, obj=%r, nest_level=%r)'%(chunklist,obj,nest_level)
        if hasattr(obj, 'json_equivalent'):
            json = self.encode_equivalent( obj, nest_level=nest_level )
            if json is not None:
                chunklist.append( json )
                return
        if obj is None:
            chunklist.append( self.encode_null() )
        elif obj is undefined:
            if self._allow_undefined_values:
                chunklist.append( self.encode_undefined() )
            else:
                raise JSONEncodeError('strict JSON does not permit "undefined" values')
        elif isinstance(obj, bool):
            chunklist.append( self.encode_boolean(obj) )
        elif isinstance(obj, (int,long,float,complex)) or \
                 (decimal and isinstance(obj, decimal.Decimal)):
            chunklist.append( self.encode_number(obj) )
        elif isinstance(obj, basestring) or isstringtype(obj):
            chunklist.append( self.encode_string(obj) )
        else:
            self.encode_composite(chunklist, obj, nest_level)

    def encode_composite(self, chunklist, obj, nest_level):
        """Encodes just dictionaries, lists, or sequences.

        Basically handles any python type for which iter() can create
        an iterator object.

        This method is not intended to be called directly.  Use the
        encode() method instead.

        """
        #print 'encode_complex_helper(chunklist=%r, obj=%r, nest_level=%r)'%(chunklist,obj,nest_level)
        try:
            # Is it a dictionary or UserDict?  Try iterkeys method first.
            it = obj.iterkeys()
        except AttributeError:
            try:
                # Is it a sequence?  Try to make an iterator for it.
                it = iter(obj)
            except TypeError:
                it = None
        if it is not None:
            # Does it look like a dictionary?  Check for a minimal dict or
            # UserDict interface.
            isdict = hasattr(obj, '__getitem__') and hasattr(obj, 'keys')
            compactly = self._encode_compactly
            if isdict:
                chunklist.append('{')
                if compactly:
                    dictcolon = ':'
                else:
                    dictcolon = ' : '
            else:
                chunklist.append('[')
            #print nest_level, 'opening sequence:', repr(chunklist)
            if not compactly:
                indent0 = '  ' * nest_level
                indent = '  ' * (nest_level+1)
                chunklist.append(' ')
            sequence_chunks = []  # use this to allow sorting afterwards if dict
            try: # while not StopIteration
                numitems = 0
                while True:
                    obj2 = it.next()
                    if obj2 is obj:
                        raise JSONEncodeError('trying to encode an infinite sequence',obj)
                    if isdict and not isstringtype(obj2):
                        # Check JSON restrictions on key types
                        if isnumbertype(obj2):
                            if not self._allow_nonstring_keys:
                                raise JSONEncodeError('object properties (dictionary keys) must be strings in strict JSON',obj2)
                        else:
                            raise JSONEncodeError('object properties (dictionary keys) can only be strings or numbers in ECMAScript',obj2)

                    # Encode this item in the sequence and put into item_chunks
                    item_chunks = []
                    self.encode_helper( item_chunks, obj2, nest_level=nest_level+1 )
                    if isdict:
                        item_chunks.append(dictcolon)
                        obj3 = obj[obj2]
                        self.encode_helper(item_chunks, obj3, nest_level=nest_level+2)

                    #print nest_level, numitems, 'item:', repr(obj2)
                    #print nest_level, numitems, 'sequence_chunks:', repr(sequence_chunks)
                    #print nest_level, numitems, 'item_chunks:', repr(item_chunks)
                    #extend_list_with_sep(sequence_chunks, item_chunks)
                    sequence_chunks.append(item_chunks)
                    #print nest_level, numitems, 'new sequence_chunks:', repr(sequence_chunks)
                    numitems += 1
            except StopIteration:
                pass

            if isdict and self._sort_dictionary_keys:
                sequence_chunks.sort()  # Note sorts by JSON repr, not original Python object
            if compactly:
                sep = ','
            else:
                sep = ',\n' + indent

            #print nest_level, 'closing sequence'
            #print nest_level, 'chunklist:', repr(chunklist)
            #print nest_level, 'sequence_chunks:', repr(sequence_chunks)
            extend_and_flatten_list_with_sep( chunklist, sequence_chunks, sep )
            #print nest_level, 'new chunklist:', repr(chunklist)

            if not compactly:
                if numitems > 1:
                    chunklist.append('\n' + indent0)
                else:
                    chunklist.append(' ')
            if isdict:
                chunklist.append('}')
            else:
                chunklist.append(']')
        else: # Can't create an iterator for the object
            json2 = self.encode_default( obj, nest_level=nest_level )
            chunklist.append( json2 )

    def encode_equivalent( self, obj, nest_level=0 ):
        """This method is used to encode user-defined class objects.

        The object being encoded should have a json_equivalent()
        method defined which returns another equivalent object which
        is easily JSON-encoded.  If the object in question has no
        json_equivalent() method available then None is returned
        instead of a string so that the encoding will attempt the next
        strategy.

        If a caller wishes to disable the calling of json_equivalent()
        methods, then subclass this class and override this method
        to just return None.
        
        """
        if hasattr(obj, 'json_equivalent') \
               and callable(getattr(obj,'json_equivalent')):
            obj2 = obj.json_equivalent()
            if obj2 is obj:
                # Try to prevent careless infinite recursion
                raise JSONEncodeError('object has a json_equivalent() method that returns itself',obj)
            json2 = self.encode( obj2, nest_level=nest_level )
            return json2
        else:
            return None

    def encode_default( self, obj, nest_level=0 ):
        """This method is used to encode objects into JSON which are not straightforward.

        This method is intended to be overridden by subclasses which wish
        to extend this encoder to handle additional types.

        """
        raise JSONEncodeError('can not encode object into a JSON representation',obj)


# ------------------------------

def encode( obj, strict=False, compactly=True, escape_unicode=False, encoding=None ):
    """Encodes a Python object into a JSON-encoded string.

    If 'strict' is set to True, then only strictly-conforming JSON
    output will be produced.  Note that this means that some types
    of values may not be convertable and will result in a
    JSONEncodeError exception.

    If 'compactly' is set to True, then the resulting string will
    have all extraneous white space removed; if False then the
    string will be "pretty printed" with whitespace and indentation
    added to make it more readable.

    If 'escape_unicode' is set to True, then all non-ASCII characters
    will be represented as a unicode escape sequence; if False then
    the actual real unicode character will be inserted.

    If no encoding is specified (encoding=None) then the output will
    either be a Python string (if entirely ASCII) or a Python unicode
    string type.

    However if an encoding name is given then the returned value will
    be a python string which is the byte sequence encoding the JSON
    value.  As the default/recommended encoding for JSON is UTF-8,
    you should almost always pass in encoding='utf8'.

    """
    import sys
    encoder = None # Custom codec encoding function
    bom = None  # Byte order mark to prepend to final output
    cdk = None  # Codec to use
    if encoding is not None:
        import codecs
        try:
            cdk = codecs.lookup(encoding)
        except LookupError:
            cdk = None

        if cdk:
            pass
        elif not cdk:
            # No built-in codec was found, see if it is something we
            # can do ourself.
            encoding = encoding.lower()
            if encoding.startswith('utf-32') or encoding.startswith('utf32') \
                   or encoding.startswith('ucs4') \
                   or encoding.startswith('ucs-4'):
                # Python doesn't natively have a UTF-32 codec, but JSON
                # requires that it be supported.  So we must decode these
                # manually.
                if encoding.endswith('le'):
                    encoder = utf32le_encode
                elif encoding.endswith('be'):
                    encoder = utf32be_encode
                else:
                    encoder = utf32be_encode
                    bom = codecs.BOM_UTF32_BE
            elif encoding.startswith('ucs2') or encoding.startswith('ucs-2'):
                # Python has no UCS-2, but we can simulate with
                # UTF-16.  We just need to force us to not try to
                # encode anything past the BMP.
                encoding = 'utf-16'
                if not escape_unicode and not callable(escape_unicode):
                   escape_unicode = lambda c: (0xD800 <= ord(c) <= 0xDFFF) or ord(c) >= 0x10000
            else:
                raise JSONEncodeError('this python has no codec for this character encoding',encoding)

    if not escape_unicode and not callable(escape_unicode):
        if encoding and encoding.startswith('utf'):
            # All UTF-x encodings can do the whole Unicode repertoire, so
            # do nothing special.
            pass
        else:
            # Even though we don't want to escape all unicode chars,
            # the encoding being used may force us to do so anyway.
            # We must pass in a function which says which characters
            # the encoding can handle and which it can't.
            def in_repertoire( c, encoding_func ):
                try:
                    x = encoding_func( c, errors='strict' )
                except UnicodeError:
                    return False
                return True
            if encoder:
                escape_unicode = lambda c: not in_repertoire(c, encoder)
            elif cdk:
                escape_unicode = lambda c: not in_repertoire(c, cdk[0])
            else:
                pass # Let the JSON object deal with it

    j = JSON( strict=strict, compactly=compactly, escape_unicode=escape_unicode )

    unitxt = j.encode( obj )
    if encoder:
        txt = encoder( unitxt )
    elif encoding is not None:
        txt = unitxt.encode( encoding )
    else:
        txt = unitxt
    if bom:
        txt = bom + txt
    return txt


def decode( txt, strict=False, encoding=None, **kw ):
    """Decodes a JSON-encoded string into a Python object.

    If 'strict' is set to True, then those strings that are not
    entirely strictly conforming to JSON will result in a
    JSONDecodeError exception.

    The input string can be either a python string or a python unicode
    string.  If it is already a unicode string, then it is assumed
    that no character set decoding is required.

    However, if you pass in a non-Unicode text string (i.e., a python
    type 'str') then an attempt will be made to auto-detect and decode
    the character encoding.  This will be successful if the input was
    encoded in any of UTF-8, UTF-16 (BE or LE), or UTF-32 (BE or LE),
    and of course plain ASCII works too.
    
    Note though that if you know the character encoding, then you
    should convert to a unicode string yourself, or pass it the name
    of the 'encoding' to avoid the guessing made by the auto
    detection, as with

        python_object = demjson.decode( input_bytes, encoding='utf8' )

    Optional keywords arguments must be of the form
        allow_xxxx=True/False
    or
        prevent_xxxx=True/False
    where each will allow or prevent the specific behavior, after the
    evaluation of the 'strict' argument.  For example, if strict=True
    then by also passing 'allow_comments=True' then comments will be
    allowed.  If strict=False then prevent_comments=True will allow
    everything except comments.
    
    """
    # Initialize the JSON object
    j = JSON( strict=strict )
    for keyword, value in kw.items():
        if keyword.startswith('allow_'):
            behavior = keyword[6:]
            allow = bool(value)
        elif keyword.startswith('prevent_'):
            behavior = keyword[8:]
            allow = not bool(value)
        else:
            raise ValueError('unknown keyword argument', keyword)
        if allow:
            j.allow(behavior)
        else:
            j.prevent(behavior)

    # Convert the input string into unicode if needed.
    if isinstance(txt,unicode):
        unitxt = txt
    else:
        if encoding is None:
            unitxt = auto_unicode_decode( txt )
        else:
            cdk = None # codec
            decoder = None
            import codecs
            try:
                cdk = codecs.lookup(encoding)
            except LookupError:
                encoding = encoding.lower()
                decoder = None
                if encoding.startswith('utf-32') \
                       or encoding.startswith('ucs4') \
                       or encoding.startswith('ucs-4'):
                    # Python doesn't natively have a UTF-32 codec, but JSON
                    # requires that it be supported.  So we must decode these
                    # manually.
                    if encoding.endswith('le'):
                        decoder = utf32le_decode
                    elif encoding.endswith('be'):
                        decoder = utf32be_decode
                    else:
                        if txt.startswith( codecs.BOM_UTF32_BE ):
                            decoder = utf32be_decode
                            txt = txt[4:]
                        elif txt.startswith( codecs.BOM_UTF32_LE ):
                            decoder = utf32le_decode
                            txt = txt[4:]
                        else:
                            if encoding.startswith('ucs'):
                                raise JSONDecodeError('UCS-4 encoded string must start with a BOM')
                            decoder = utf32be_decode # Default BE for UTF, per unicode spec
                elif encoding.startswith('ucs2') or encoding.startswith('ucs-2'):
                    # Python has no UCS-2, but we can simulate with
                    # UTF-16.  We just need to force us to not try to
                    # encode anything past the BMP.
                    encoding = 'utf-16'

            if decoder:
                unitxt = decoder(txt)
            elif encoding:
                unitxt = txt.decode(encoding)
            else:
                raise JSONDecodeError('this python has no codec for this character encoding',encoding)

        # Check that the decoding seems sane.  Per RFC 4627 section 3:
        #    "Since the first two characters of a JSON text will
        #    always be ASCII characters [RFC0020], ..."
        #
        # This check is probably not necessary, but it allows us to
        # raise a suitably descriptive error rather than an obscure
        # syntax error later on.
        #
        # Note that the RFC requirements of two ASCII characters seems
        # to be an incorrect statement as a JSON string literal may
        # have as it's first character any unicode character.  Thus
        # the first two characters will always be ASCII, unless the
        # first character is a quotation mark.  And in non-strict
        # mode we can also have a few other characters too.
        if len(unitxt) > 2:
            first, second = unitxt[:2]
            if first in '"\'':
                pass # second can be anything inside string literal
            else:
                if ((ord(first) < 0x20 or ord(first) > 0x7f) or \
                    (ord(second) < 0x20 or ord(second) > 0x7f)) and \
                    (not j.isws(first) and not j.isws(second)):
                    # Found non-printable ascii, must check unicode
                    # categories to see if the character is legal.
                    # Only whitespace, line and paragraph separators,
                    # and format control chars are legal here.
                    import unicodedata
                    catfirst = unicodedata.category(unicode(first))
                    catsecond = unicodedata.category(unicode(second))
                    if catfirst not in ('Zs','Zl','Zp','Cf') or \
                           catsecond not in ('Zs','Zl','Zp','Cf'):
                        raise JSONDecodeError('the decoded string is gibberish, is the encoding correct?',encoding)
    # Now ready to do the actual decoding
    obj = j.decode( unitxt )
    return obj

# end file



#!/usr/bin/env python
from setuptools import setup, find_packages
import os

here = os.path.abspath(os.path.dirname(__file__))
README = open(os.path.join(here, 'README.rst')).read()
NEWS = open(os.path.join(here, 'NEWS.txt')).read()


version = '0.3.0'


setup(name='stupeflix_api',
    version=version,
    description="Stupeflix web services client libraries",
    long_description=README + '\n\n' + NEWS,
    classifiers=[
      # Get strings from http://pypi.python.org/pypi?%3Aaction=list_classifiers
    ],
    keywords='',
    author='Francois Lagunas',
    author_email='francois@stupeflix.com',
    url='https://github.com/stupeflix/Stupeflix-API-Client',
    license='',
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
    entry_points={
        'console_scripts':
            ['stupeflix-api=stupeflix_api.commandline:main']
    }
)



import os
import time

# These are optional variables, by default read from the environement variables, but you can too override them with your own credentials directly.
youtubeLogin = os.getenv("YOUTUBE_LOGIN")
youtubePassword = os.getenv("YOUTUBE_PASSWORD")
youtubeDeveloperKey = os.getenv("YOUTUBE_DEVELOPER_KEY")
youtubeAuthSubToken= os.getenv("YOUTUBE_AUTH_SUB_TOKEN")
dailymotionLogin = os.getenv("DAILYMOTION_LOGIN")
dailymotionPassword = os.getenv("DAILYMOTION_PASSWORD")
s3AccessKey = os.getenv("S3_ACCESS_KEY")
s3SecretKey = os.getenv("S3_SECRET_KEY")
s3Bucket = os.getenv("S3_BUCKET")
httpUploadPrefix = os.getenv("HTTP_UPLOAD_PREFIX")

if os.getenv("STUPEFLIX_TEST_TIME") == None:
    t = time.time()
    dateString = time.strftime("%Ya%ma%da%Ha%Ma%S", time.gmtime(int(t)))
else:
    dateString = os.getenv("STUPEFLIX_TEST_TIME")

filename = os.getenv("STUPEFLIX_MOVIE")
if filename == None:
    filename = "movie.xml"

facebookApiKey = os.getenv("FACEBOOK_API_KEY")
facebookApiSecret = os.getenv("FACEBOOK_API_SECRET")
facebookToken = os.getenv("FACEBOOK_TOKEN")



import unittest
import logging
import mox
import mock
from nose.tools import assert_equal
from stupeflix_api.connection import Connection
from stupeflix_api import httplib2


class ConnectionTests(unittest.TestCase):

    def setUp(self):
        self.mox = mox.Mox()

    def tearDown(self):
        self.mox.UnsetStubs()

    def disable_retries(self):
        self.orig_max_retries = Connection.MAX_NETWORK_RETRY
        Connection.MAX_NETWORK_RETRY = 1

    def restore_retries(self):
        Connection.MAX_NETWORK_RETRY = self.orig_max_retries 

    def test_request(self):
        self.disable_retries()

        # Setup mocks
        http = self.mox.CreateMock(httplib2.Http)
        http.request(u'http://my.site.com', 'GET', body=None,
                headers={'User-Agent': 'Basic Agent'},
                sendcallback=None).AndReturn(("headers", "content"))

        orig_http = httplib2.Http
        httplib2.Http = mock.Mock(return_value=http)

        # Test the request method
        self.mox.ReplayAll()
        connection = Connection("http://my.site.com/")
        ret = connection.request("foo")
        self.assertEqual(ret, {"body": "content", "headers": "headers"})

        # Verify mocks
        self.mox.VerifyAll()
        httplib2.Http = orig_http

        self.restore_retries()

    def test_request_raw(self):
        self.disable_retries()

        # Setup mocks
        http = self.mox.CreateMock(httplib2.Http)
        http.request('http://my.site.com/', 'GET', body=None, headers={},
                sendcallback=None).AndReturn(("headers", "content"))

        orig_http = httplib2.Http
        httplib2.Http = mock.Mock(return_value=http)

        # Test the request method
        self.mox.ReplayAll()
        connection = Connection("http://my.site.com/")
        ret = connection.request_raw()
        self.assertEqual(ret, {"body": "content", "headers": "headers"})

        # Verify mocks
        self.mox.VerifyAll()
        httplib2.Http = orig_http

        self.restore_retries()

    def test_request_errors(self):
        # Setup mocks
        http = self.mox.CreateMock(httplib2.Http)
        for i in range(Connection.MAX_NETWORK_RETRY):
            http.request(u'http://my.site.com', 'GET', body=None,
                    headers={'User-Agent': 'Basic Agent'},
                    sendcallback=None).AndRaise(httplib2.ServerNotFoundError)

        orig_http = httplib2.Http
        httplib2.Http = mock.Mock(return_value=http)

        logger = logging.getLogger("stupeflix_api.connection")
        handler = TestHandler()
        logger.addHandler(handler)

        # Test the request method
        self.mox.ReplayAll()
        connection = Connection("http://my.site.com/")
        self.assertRaises(httplib2.ServerNotFoundError, 
                connection.request, "foo")
        handler.assert_handled_count(4)

        # Verify mocks
        self.mox.VerifyAll()
        httplib2.Http = orig_http


class TestHandler(logging.Handler):

    def __init__(self, *args, **kwargs):
        super(TestHandler, self).__init__(*args, **kwargs)
        self.recorder = mock.Mock()

    def emit(self, record):
        self.recorder(record)

    def assert_handled_count(self, num_times):
        assert_equal(self.recorder.call_count, num_times)



import urlparse
import urllib
import time
import os.path
import os
import logging
from stupeflix_api import httplib2


httplib2.debuglevel = 0
logger = logging.getLogger(__name__)


class Connection(object):

    MAX_NETWORK_RETRY = 5

    def __init__(self, base_url, username=None, password=None, followRedirect = True, sendHTTP100Continue = True, userAgent = 'Basic Agent', timeout = None):
        self.base_url = base_url
        self.username = username
        self.userAgent = userAgent
        self.timeout = timeout
        self.url = urlparse.urlparse(base_url)
        
        (scheme, netloc, path, query, fragment) = urlparse.urlsplit(base_url)
            
        self.scheme = scheme
        self.host = netloc
        self.path = path
        
        self.followRedirect = followRedirect
        self.username = username
        self.password = password

        self.debuglevel = 0
        self.sendHTTP100Continue = sendHTTP100Continue
        self.createHttp()

    def createHttp(self):
        self.h = httplib2.Http(cache = None, timeout = self.timeout)
        self.h.follow_redirects = self.followRedirect
        self.h.follow_all_redirects = self.followRedirect
        self.h.sendHTTP100Continue = self.sendHTTP100Continue

        # Create Http class with support for Digest HTTP Authentication, if necessary
        if self.username and self.password:
            self.h.add_credentials(self.username, self.password)
    
    def request_get(self, resource, args = None, headers={}):
        return self.request(resource, "get", args, headers=headers)
        
    def request_delete(self, resource, args = None, headers={}):
        return self.request(resource, "delete", args, headers=headers)
        
    def request_head(self, resource, args = None, headers={}):
        return self.request(resource, "head", args, headers=headers)
        
    def request_post(self, resource, args = None, body = None, filename=None, headers={}):        
        return self.request(resource, "post", args , body = body, filename=filename, headers=headers)
        
    def request_put(self, resource, args = None, body = None, filename=None, headers={}, sendcallback = None):
        resp = self.request(resource, "put", args , body = body, filename=filename, headers=headers, sendcallback = sendcallback)
        return resp
 
    def dump(self, message, request_uri, method):
        if self.debuglevel > 0:
            print message, " " , time.asctime(), " " , request_uri, "\n"

    def request(self, resource, method = "get", args = None, body = None, filename=None, headers={}, sendcallback = None):        
        path = resource

        if headers == None:
            headers = {}

        headers['User-Agent'] = self.userAgent
        
        # TEMPORARY : add support for streaming
        if method != "get":
            if not body and filename:
                body = open(filename, 'rb')
                if not "Content-Length" in headers:
                    headers["Content-Length"] = str(os.stat(filename).st_size)


        if args:
            path += u"?" + urllib.urlencode(args)
            
        request_path = []
        if self.path != "/":
            if self.path.endswith('/'):
                request_path.append(self.path[:-1])
            else:
                request_path.append(self.path)
            if path.startswith('/'):
                request_path.append(path[1:])
            else:
                request_path.append(path)
                
        uri = u"%s://%s%s" % (self.scheme, self.host, u'/'.join(request_path))

        for i in range(self.MAX_NETWORK_RETRY):
            try:
                resp, content = self.h.request(uri, method.upper(), body=body, headers=headers, sendcallback = sendcallback)
                break
            except:
                if i < self.MAX_NETWORK_RETRY - 1:
                    logger.info("request failed, retrying (%s/%s)", i + 1,
                            self.MAX_NETWORK_RETRY, exc_info=True)
                    self.createHttp()
                else:
                    raise
        
        return {u'headers':resp, u'body':content}

    def request_raw(self, method = "get", args = None, body = None, filename=None, headers={}, sendcallback = None):
        if body == None and filename != None:
            body = open(filename, 'rb')

        for i in range(self.MAX_NETWORK_RETRY):
            try:
                resp, content = self.h.request(self.base_url, method.upper(), body=body, headers=headers, sendcallback = sendcallback )                
                break
            except:
                if i < self.MAX_NETWORK_RETRY - 1:
                    logger.info("raw request failed, retrying (%s/%s)", i + 1,
                            self.MAX_NETWORK_RETRY, exc_info=True)
                    self.createHttp()
                else:
                    raise

        return {u'headers':resp, u'body':content}
